/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Fabien Badeig, Quentin Perlorson
 * Contacts :
 * perception-naolab@inria.fr
 *
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
* @file commondefinitions.h
* @author Fabien Badeig, Pelorson Quentin
* @date 2015 Feb 25
* @brief Memory organization with structures
**/

#ifndef SHAREDMEMORYDEFINITIONS_H
#define SHAREDMEMORYDEFINITIONS_H

#include <string>
#include <boost/interprocess/sync/scoped_lock.hpp>
#include <boost/interprocess/shared_memory_object.hpp>
#include <boost/interprocess/mapped_region.hpp>
#include <boost/interprocess/sync/interprocess_mutex.hpp>
#include <boost/interprocess/sync/interprocess_condition.hpp>
#include <boost/interprocess/sync/interprocess_condition_any.hpp>
#include <boost/interprocess/sync/interprocess_upgradable_mutex.hpp>
#include <boost/interprocess/sync/sharable_lock.hpp>
#include <boost/interprocess/sync/upgradable_lock.hpp>

/* buffer id to manage sound buffer */
static std::string shared_memory_audio_name = "NaoRemoteDriverAudioSharedMemory";
/* buffer id to store NAO posture information */
static std::string shared_memory_posture_name = "NaoRemoteDriverPostureSharedMemory";
/* buffer id to store results of face detector from NAO */
static std::string shared_memory_facedetector_name = "NaoRemoteDriverFaceDetectorSharedMemory";
/* Header for motor management (control motor, get info about motor) */
static std::string shared_memory_motion_control_name = "NaoRemoteDriverMotionControlSharedMemory";
/* Content about current motor positions */
static std::string shared_memory_motion_contentg_name = "NaoRemoteDriverMotionContentPSharedMemory";
/* Content about motor control */
static std::string shared_memory_motion_contentp_name = "NaoRemoteDriverMotionContentGSharedMemory";
static std::string shared_memory_move_name = "NaoRemoteDriverMoveSharedMemory";
/* Save for compatibility with previous modules */
static std::string shared_memory_head_motion_name = "NaoRemoteDriverHeadMotionSharedMemory";
/* Two image buffers to store left and right camera image */
static std::string shared_memory_video0_name = "NaoRemoteDriverVideo0SharedMemory";
static std::string shared_memory_video1_name = "NaoRemoteDriverVideo1SharedMemory";

/* Nao motor ID*/
enum NAO_partID {
    HEAD_YAW = 0,
    HEAD_PITCH,
    L_SHOULDER_PITCH,
    L_SHOULDER_ROLL,
    L_ELBOW_YAW,
    L_ELBOW_ROLL,
    L_WRIST_YAW,
    L_HAND,
    L_HIP_YAW_PITCH,
    L_HIP_ROLL,
    L_HIP_PITCH,
    L_KNEE_PITCH,
    L_ANKLE_PITCH,
    L_ANKLE_ROLL,
    R_SHOULDER_PITCH,
    R_SHOULDER_ROLL,
    R_ELBOW_YAW,
    R_ELBOW_ROLL,
    R_WRIST_YAW,
    R_HAND,
    R_HIP_YAW_PITCH,
    R_HIP_ROLL,
    R_HIP_PITCH,
    R_KNEE_PITCH,
    R_ANKLE_PITCH,
    R_ANKLE_ROLL
};

/* Info about Nao motor*/
struct NAO_part {
    int ID;
    std::string name;
    std::string part;
    float min_range;
    float max_range;
    float min_collision_angle;
    float max_collision_angle;
};

/* Nao motor specifications*/
static const struct NAO_part NAO_control_table[26] = {
    // ID                       //NAME          //PART  // RANGE MIN // RANGE MAX // MIN VALUE // MAX VALUE
    {HEAD_YAW,         "HeadYaw",          "Head", -2.085f,    2.085f,     -2.085f,    2.085f},
    {HEAD_PITCH,       "HeadPitch",        "Head", -0.672f,    0.5149f,    -0.33f,     0.2f},
    {L_SHOULDER_PITCH, "LShoulderPitch",   "LArm", -2.085f,    2.085f,     -2.085f,    2.085f},
    {L_SHOULDER_ROLL,  "LShoulderRoll",    "LArm", -0.3142f,   1.3265f,    -0.3142f,   1.3265f},
    {L_ELBOW_YAW,      "LElbowYaw",        "LArm", -2.085f,    2.085f,     -2.085f,    2.085f},
    {L_ELBOW_ROLL,     "LElbowRoll",       "LArm", -1.5446f,   -0.0349f,   -1.5446f,   -0.0349f},
    {L_WRIST_YAW,      "LWristYaw",        "LArm", -1.8238f,   1.8238f,    -1.8238f,   1.8238f},
    {L_HAND,           "LHand",            "LArm", 0.0f,       1.0f,       0.0f,       1.0f},
    {L_HIP_YAW_PITCH,  "LHipYawPitch",     "LLeg", -1.1453f,   0.74f,      -1.1453f,   0.74f},
    {L_HIP_ROLL,       "LHipRoll",         "LLeg", -0.3794f,   0.79f,      -0.3794f,   0.79f},
    {L_HIP_PITCH,      "LHipPitch",        "LLeg", -1.5358f,   0.484f,     -1.5358f,   0.484f},
    {L_KNEE_PITCH,     "LKneePitch",       "LLeg", -0.092f,    2.1125f,    -0.092f,    2.1125f},
    {L_ANKLE_PITCH,    "LAnklePitch",      "LLeg", -1.1895f,   0.9227f,    -1.1895f,   0.9227f},
    {L_ANKLE_ROLL,     "LAnkleRoll",       "LLeg", -0.3978f,   0.769f,     -0.0f,      0.0499f},
    {R_SHOULDER_PITCH, "RShoulderPitch",   "RArm", -2.085f,    2.085f,     -2.085f,    2.085f},
    {R_SHOULDER_ROLL,  "RShoulderRoll",    "RArm", -1.3265f,   0.3142f,    -1.3265f,   0.3142f},
    {R_ELBOW_YAW,      "RElbowYaw",        "RArm", -2.085f,    2.085f,     -2.085f,    2.085f},
    {R_ELBOW_ROLL,     "RElbowRoll",       "RArm", 0.0349f,    1.5446f,    0.0349f,    1.5446f},
    {R_WRIST_YAW,      "RWristYaw",        "RArm", -1.8238,    1.8238f,    -1.8238,    1.8238f},
    {R_HAND,           "RHand",            "RArm", 0.0f,       1.0f,       0.0f,       1.0f},
    {R_HIP_YAW_PITCH,  "RHipYawPitch",     "RLeg", -1.1453f,   0.74f,      -1.1453f,   0.74f},
    {R_HIP_ROLL,       "RHipRoll",         "RLeg", -0.79f,     0.3794f,    -0.79f,     0.3794f},
    {R_HIP_PITCH,      "RHipPitch",        "RLeg", -1.5358f,   0.484f,     -1.5358f,   0.484f},
    {R_KNEE_PITCH,     "RKneePitch",       "RLeg", -0.103f,    2.12f,      -0.103f,    2.12f},
    {R_ANKLE_PITCH,    "RAnklePitch",      "RLeg", -1.1864f,   0.932f,     -1.1864f,   0.932f},
    {R_ANKLE_ROLL,     "RAnkleRoll",       "RLeg", -0.7689f,   0.3979f,    -0.0499f,   0.0f},
};

/* Buffer to store image data */
struct ImageMessage {

    ImageMessage()
    {}

    //Mutex to protect access to the conds
    //boost::interprocess::interprocess_mutex bufferMutex;
    boost::interprocess::interprocess_upgradable_mutex imageMutex;

    //Condition to wait
    boost::interprocess::interprocess_condition_any imageCond;

    int ts;
    int tms;
    int width;
    int height;
    int nbChannels;

    //Image data
    unsigned char items[640 * 480 * 3];
};

/* Buffer to store audio data */
struct AudioMessage {
    AudioMessage()
    {}

    //Mutex to protect access to the conds
    //boost::interprocess::interprocess_mutex bufferMutex;
    boost::interprocess::interprocess_upgradable_mutex soundMutex;

    //Condition to wait
    boost::interprocess::interprocess_condition_any soundCond;

    int ts;
    int tms;
    int samples;
    int channels;
    int fRate;

    short data[16384];
};

/* Buffer to manage posture state */
struct PostureMessage {

    PostureMessage()
    {
    }
    //Mutex to protect access to the conds
    boost::interprocess::interprocess_mutex newCommandMutex;

    //Condition to wait
    boost::interprocess::interprocess_condition  newCommandCond;

    //Condition to wait
    boost::interprocess::interprocess_condition  motionFinishedCond;

    char type[11]; //predefined posture
    float speed;
};

/* motion header */
struct MotionMessageControl {

    MotionMessageControl()
    {
    }
    //Mutex to protect access to the conds
    boost::interprocess::interprocess_mutex newCommandMutex;

    //Condition to wait
    boost::interprocess::interprocess_condition  newCommandCond;

    //Condition to wait
    boost::interprocess::interprocess_condition  motionFinishedCond;

    char type[3]; //get put
    int jointArrayLength; //get -1 ==> all
};

/* motion control */
struct MotionMessageContentP {

    MotionMessageContentP()
    {
    }

    float angleLists[26];
    float angleSpeeds[26];
    bool mask[26];
};

/* motion access */
struct MotionMessageContentG {

    MotionMessageContentG()
    {
    }

    float currentPositions[26];
    bool mask[26];
};

/* NAO walk control */
struct MoveMessage {
    MoveMessage()
    {
    }

    //Mutex to protect access to the conds
    boost::interprocess::interprocess_mutex newCommandMutex;

    //Condition to wait
    boost::interprocess::interprocess_condition  newCommandCond;

    //Condition to wait
    boost::interprocess::interprocess_condition  motionFinishedCond;

    float xPos;
    float yPos;
    float theta;
};


struct HeadMotionMessage {

    HeadMotionMessage()
    {}

    //Mutex to protect access to the conds
    boost::interprocess::interprocess_mutex newCommandMutex;

    //Condition to wait
    boost::interprocess::interprocess_condition  newCommandCond;

    //Condition to wait
    boost::interprocess::interprocess_condition  motionFinishedCond;

    bool absoluteMove;

    //Items to fill
    float items[4];
};

struct FaceDetectorMessage {

    FaceDetectorMessage(): nbFaces(0)
    {}

    //Mutex to protect access to the conds
    //boost::interprocess::interprocess_mutex faceDetectorMutex;
    boost::interprocess::interprocess_upgradable_mutex fdetectMutex;

    //Condition to wait
    //boost::interprocess::interprocess_condition  newDataCond;
    boost::interprocess::interprocess_condition_any fdetectCond;

    int ts;
    int tms;
    int nbFaces;

    //Items to fill
    float items[2 * 50];
};

#endif // SHAREDMEMORYDEFINITIONS_H

