/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Quentin Perlorson, Fabien Badeig
 * Contacts :
 * perception-naolab@inria.fr
 * 
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
* @file dataproxies.cpp
* @author Pelorson Quentin, Fabien Badeig
* @date 2015 Feb 25
* @brief
**/

#include "dataproxies.h"
#include "commondefinitions.h"
#include <stdexcept>

using namespace boost::interprocess;

void DataProxy::moveHead(float yawAngle, float pitchAngle, bool absolute, float &newAbsoluteYaw, float newAbsolutePitch)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm(open_only , shared_memory_head_motion_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region(shm , read_write);

        void *addr = region.get_address();
        HeadMotionMessage *data = static_cast<HeadMotionMessage *>(addr);
        data->items[0] = yawAngle;
        data->items[1] = pitchAngle;
        data->absoluteMove = absolute;
        scoped_lock<interprocess_mutex> lock(data->newCommandMutex);
        data->newCommandCond.notify_one();
        data->motionFinishedCond.wait(lock);
        newAbsoluteYaw = data->items[2];
        newAbsolutePitch = data->items[3];

    }
    catch (interprocess_exception &ex) {
        std::cout << "moveHead: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::getAllMotorInfo( float *currentAngles )
{
    try {
        //Open the shared memory object.
        shared_memory_object shm0(open_only , shared_memory_motion_control_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region0(shm0 , read_write);

        void *addr0 = region0.get_address();
        MotionMessageControl *data = static_cast<MotionMessageControl *>(addr0);

        //Open the shared memory object.
        shared_memory_object shmG(open_only , shared_memory_motion_contentg_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region regionG(shmG , read_write);

        void *addrG = regionG.get_address();
        MotionMessageContentG *contentG = static_cast<MotionMessageContentG *>(addrG);

        scoped_lock<interprocess_mutex> lock(data->newCommandMutex);
        data->jointArrayLength = -1;
        data->type[0] = 'g';
        data->type[1] = 'e';
        data->type[2] = 't';

        for (int i = 0; i < 26; i++) {
            contentG->mask[i] = true;
        }

        data->newCommandCond.notify_one();
        data->motionFinishedCond.wait(lock);

        for (int i = 0; i < 26; i++) {
            currentAngles[i] = contentG->currentPositions[i];
        }


    }
    catch (interprocess_exception &ex) {
        std::cout << "getAllMotorInfo: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::getMotorInfo(double *lJoints, int nbElt, float *currentAngles)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm0(open_only , shared_memory_motion_control_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region0(shm0 , read_write);

        void *addr0 = region0.get_address();
        MotionMessageControl *data = static_cast<MotionMessageControl *>(addr0);

        //Open the shared memory object.
        shared_memory_object shmG(open_only , shared_memory_motion_contentg_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region regionG(shmG , read_write);

        void *addrG = regionG.get_address();
        MotionMessageContentG *contentG = static_cast<MotionMessageContentG *>(addrG);

        scoped_lock<interprocess_mutex> lock(data->newCommandMutex);
        data->jointArrayLength = nbElt;
        data->type[0] = 'g';
        data->type[1] = 'e';
        data->type[2] = 't';

        for (int i = 0; i < 26; i++) {
            contentG->mask[i] = false;
        }


        for (int i = 0; i < nbElt; i++) {
            int a =  (int)lJoints[i];
            contentG->mask[a] = true;
        }

        data->newCommandCond.notify_one();
        data->motionFinishedCond.wait(lock);

        for (int i = 0; i < nbElt; i++) {
            int a =  (int)lJoints[i];
            currentAngles[i] = contentG->currentPositions[a];
        }
    }
    catch (interprocess_exception &ex) {
        std::cout << "getMotorInfo: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::setNaoPosture(std::string &posture, float &speed)
{
    std::cerr << posture << std::endl;
    try {
        //Open the shared memory object.
        shared_memory_object shm(open_only , shared_memory_posture_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region(shm , read_write);

        void *addr = region.get_address();
        PostureMessage *data = static_cast<PostureMessage *>(addr);

        scoped_lock<interprocess_mutex> lock(data->newCommandMutex);

        int i = 0;
        for (; i < posture.size(); i++) {
            data->type[i] = posture.at(i);
        }
        data->type[i] = '\0';

        data->speed = speed;

        data->newCommandCond.notify_one();
        data->motionFinishedCond.wait(lock);

    }
    catch (interprocess_exception &ex) {
        std::cout << "setNaoPosture: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::moveToPoint(float &x, float &y, float &theta)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm(open_only , shared_memory_move_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region(shm , read_write);

        void *addr = region.get_address();
        MoveMessage *data = static_cast<MoveMessage *>(addr);

        scoped_lock<interprocess_mutex> lock(data->newCommandMutex);

        data->xPos = x;
        data->yPos = y;
        data->theta = theta;

        data->newCommandCond.notify_one();
        data->motionFinishedCond.wait(lock);

    }
    catch (interprocess_exception &ex) {
        std::cout << "moveToPoint: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::moveMotorTo(double *lJoints, double *lAngles, double *lSpeeds, int nbElt, float *currentAngles)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm0(open_only , shared_memory_motion_control_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region0(shm0 , read_write);

        void *addr0 = region0.get_address();
        MotionMessageControl *data = static_cast<MotionMessageControl *>(addr0);

        //Open the shared memory object.
        shared_memory_object shmG(open_only , shared_memory_motion_contentg_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region regionG(shmG , read_write);

        void *addrG = regionG.get_address();
        MotionMessageContentG *contentG = static_cast<MotionMessageContentG *>(addrG);

        //Open the shared memory object.
        shared_memory_object shmP(open_only , shared_memory_motion_contentp_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region regionP(shmP , read_write);

        void *addrP = regionP.get_address();
        MotionMessageContentP *contentP = static_cast<MotionMessageContentP *>(addrP);

        scoped_lock<interprocess_mutex> lock(data->newCommandMutex);
        data->jointArrayLength = nbElt;
        data->type[0] = 'p';
        data->type[1] = 'u';
        data->type[2] = 't';

        for (int i = 0; i < 26; i++) {
            contentP->mask[i] = false;
        }


        for (int i = 0; i < nbElt; i++) {
            int a =  (int)lJoints[i];
            contentP->angleLists[a] = (float)lAngles[i];
            contentP->angleSpeeds[a] = (float)lSpeeds[i];
            contentP->mask[a] = true;
        }

        data->newCommandCond.notify_one();
        data->motionFinishedCond.wait(lock);

        for (int i = 0; i < nbElt; i++) {
            int a =  (int)lJoints[i];
            currentAngles[i] = contentG->currentPositions[a];
        }

    }
    catch (interprocess_exception &ex) {
        std::cout << "moveMotorTo: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::getFaceDetectorData(int &ts, int &tms, int &nbFaces, std::vector<float> &facesAngles)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm(open_only , shared_memory_facedetector_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region(shm , read_write);

        void *addr = region.get_address();
        FaceDetectorMessage *data = static_cast<FaceDetectorMessage *>(addr);

        boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock(data->fdetectMutex);

        ts = data->ts;
        tms = data->tms;
        nbFaces = data->nbFaces;

        for (int i = 0 ; i < 2 * data->nbFaces ; ++i) {
            facesAngles.push_back(data->items[i]);
        }

    }
    catch (interprocess_exception &ex) {
        std::cout << "getFaceDetectorData: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::getFaceDetectorSyncData(int &ts, int &tms, int &nbFaces, std::vector<float> &facesAngles)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm(open_only , shared_memory_facedetector_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region(shm , read_write);

        void *addr = region.get_address();
        FaceDetectorMessage *data = static_cast<FaceDetectorMessage *>(addr);

        boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock(data->fdetectMutex);
        data->fdetectCond.wait(lock);

        ts = data->ts;
        tms = data->tms;
        nbFaces = data->nbFaces;

        for (int i = 0 ; i < 2 * data->nbFaces ; ++i) {
            facesAngles.push_back(data->items[i]);
        }

    }
    catch (interprocess_exception &ex) {
        std::cout << "getFaceDetectorSyncData: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::getImages(int &ts0, int &tms0, int &width0, int &height0, int &nbChnls0, unsigned char *imgBuffer0, int &ts1, int &tms1, int &width1, int &height1, int &nbChnls1, unsigned char *imgBuffer1, int imageFormat)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm0(open_only , shared_memory_video0_name.c_str(), read_write);
        shared_memory_object shm1(open_only , shared_memory_video1_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region0(shm0 , read_write);
        mapped_region region1(shm1 , read_write);

        //Get a pointer to this shared memory block.
        void *addr0 = region0.get_address();
        ImageMessage *data0 = static_cast<ImageMessage *>(addr0);

        {
            boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock0(data0->imageMutex);
            data0->imageCond.wait(lock0);
            ts0 = data0->ts;
            tms0 = data0->tms;
            width0 = data0->width;
            height0 = data0->height;
            nbChnls0 = data0->nbChannels;

            //Read the data buffer.
            unsigned char *bufferPtr0 = data0->items;

            switch (imageFormat) {
                case RGB :
                    for (int i = 0 ; i < width0 * height0 * nbChnls0 ; ++i) {
                        imgBuffer0[i] = bufferPtr0[i];
                    }
                    break;
                case BGR :
                    for (int i = 0 ; i < width0 * height0 ; ++i) {
                        imgBuffer0[3 * i] = bufferPtr0[3 * i + 2];
                        imgBuffer0[3 * i + 1] = bufferPtr0[3 * i + 1];
                        imgBuffer0[3 * i + 2] = bufferPtr0[3 * i];
                    }
                    break;
                case MATLABRGB :
                    for (int i = 0 ; i < width0 ; ++i)
                        for (int j = 0 ; j < height0 ; ++j) {
                            imgBuffer0[i * height0 + j] = bufferPtr0[3 * j * width0 + 3 * i];
                            imgBuffer0[width0 * height0 + i * height0 + j] = bufferPtr0[3 * j * width0 + 3 * i + 1];
                            imgBuffer0[2 * width0 * height0 + i * height0 + j] = bufferPtr0[3 * j * width0 + 3 * i + 2];
                        }
                    break;
                default :
                    for (int i = 0 ; i < width0 * height0 * nbChnls0 ; ++i) {
                        imgBuffer0[i] = bufferPtr0[i];
                    }
                    break;
            }
        }

        void *addr1 = region1.get_address();
        ImageMessage *data1 = static_cast<ImageMessage *>(addr1);

        {
            boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock1(data1->imageMutex);
            data1->imageCond.wait(lock1);
            ts1 = data1->ts;
            tms1 = data1->tms;
            width1 = data1->width;
            height1 = data1->height;
            nbChnls1 = data1->nbChannels;

            //Read the data buffer.
            unsigned char *bufferPtr1 = data1->items;

            switch (imageFormat) {
                case RGB :
                    for (int i = 0 ; i < width1 * height1 * nbChnls1 ; ++i) {
                        imgBuffer1[i] = bufferPtr1[i];
                    }
                    break;
                case BGR :
                    for (int i = 0 ; i < width1 * height1 ; ++i) {
                        imgBuffer1[3 * i] = bufferPtr1[3 * i + 2];
                        imgBuffer1[3 * i + 1] = bufferPtr1[3 * i + 1];
                        imgBuffer1[3 * i + 2] = bufferPtr1[3 * i];
                    }
                    break;
                case MATLABRGB :
                    for (int i = 0 ; i < width1 ; ++i)
                        for (int j = 0 ; j < height1 ; ++j) {
                            imgBuffer1[i * height1 + j] = bufferPtr1[3 * j * width1 + 3 * i];
                            imgBuffer1[width1 * height1 + i * height1 + j] = bufferPtr1[3 * j * width1 + 3 * i + 1];
                            imgBuffer1[2 * width1 * height1 + i * height1 + j] = bufferPtr1[3 * j * width1 + 3 * i + 2];
                        }
                    break;
                default :
                    for (int i = 0 ; i < width1 * height1 * nbChnls1 ; ++i) {
                        imgBuffer1[i] = bufferPtr1[i];
                    }
                    break;
            }
        }

    }
    catch (interprocess_exception &ex) {
        std::cout << "getImages: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }

}

void DataProxy::getSyncImages(int &ts0, int &tms0, int &width0, int &height0, int &nbChnls0, unsigned char *imgBuffer0, int &ts1, int &tms1, int &width1, int &height1, int &nbChnls1, unsigned char *imgBuffer1, int imageFormat)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm0(open_only , shared_memory_video0_name.c_str(), read_write);
        shared_memory_object shm1(open_only , shared_memory_video1_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region0(shm0 , read_write);
        mapped_region region1(shm1 , read_write);

        //Get a pointer to this shared memory block.
        void *addr0 = region0.get_address();
        ImageMessage *data0 = static_cast<ImageMessage *>(addr0);

        {
            boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock0(data0->imageMutex);
            data0->imageCond.wait(lock0);
            ts0 = data0->ts;
            tms0 = data0->tms;
            width0 = data0->width;
            height0 = data0->height;
            nbChnls0 = data0->nbChannels;

            //Read the data buffer.
            unsigned char *bufferPtr0 = data0->items;

            switch (imageFormat) {
                case RGB :
                    for (int i = 0 ; i < width0 * height0 * nbChnls0 ; ++i) {
                        imgBuffer0[i] = bufferPtr0[i];
                    }
                    break;
                case BGR :
                    for (int i = 0 ; i < width0 * height0 ; ++i) {
                        imgBuffer0[3 * i] = bufferPtr0[3 * i + 2];
                        imgBuffer0[3 * i + 1] = bufferPtr0[3 * i + 1];
                        imgBuffer0[3 * i + 2] = bufferPtr0[3 * i];
                    }
                    break;
                case MATLABRGB :
                    for (int i = 0 ; i < width0 ; ++i)
                        for (int j = 0 ; j < height0 ; ++j) {
                            imgBuffer0[i * height0 + j] = bufferPtr0[3 * j * width0 + 3 * i];
                            imgBuffer0[width0 * height0 + i * height0 + j] = bufferPtr0[3 * j * width0 + 3 * i + 1];
                            imgBuffer0[2 * width0 * height0 + i * height0 + j] = bufferPtr0[3 * j * width0 + 3 * i + 2];
                        }
                    break;
                default :
                    for (int i = 0 ; i < width0 * height0 * nbChnls0 ; ++i) {
                        imgBuffer0[i] = bufferPtr0[i];
                    }
                    break;
            }
        }

        void *addr1 = region1.get_address();
        ImageMessage *data1 = static_cast<ImageMessage *>(addr1);

        {
            boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock1(data1->imageMutex);
            data1->imageCond.wait(lock1);
            ts1 = data1->ts;
            tms1 = data1->tms;
            width1 = data1->width;
            height1 = data1->height;
            nbChnls1 = data1->nbChannels;

            //Read the data buffer.
            unsigned char *bufferPtr1 = data1->items;

            switch (imageFormat) {
                case RGB :
                    for (int i = 0 ; i < width1 * height1 * nbChnls1 ; ++i) {
                        imgBuffer1[i] = bufferPtr1[i];
                    }
                    break;
                case BGR :
                    for (int i = 0 ; i < width1 * height1 ; ++i) {
                        imgBuffer1[3 * i] = bufferPtr1[3 * i + 2];
                        imgBuffer1[3 * i + 1] = bufferPtr1[3 * i + 1];
                        imgBuffer1[3 * i + 2] = bufferPtr1[3 * i];
                    }
                    break;
                case MATLABRGB :
                    for (int i = 0 ; i < width1 ; ++i)
                        for (int j = 0 ; j < height1 ; ++j) {
                            imgBuffer1[i * height1 + j] = bufferPtr1[3 * j * width1 + 3 * i];
                            imgBuffer1[width1 * height1 + i * height1 + j] = bufferPtr1[3 * j * width1 + 3 * i + 1];
                            imgBuffer1[2 * width1 * height1 + i * height1 + j] = bufferPtr1[3 * j * width1 + 3 * i + 2];
                        }
                    break;
                default :
                    for (int i = 0 ; i < width1 * height1 * nbChnls1 ; ++i) {
                        imgBuffer1[i] = bufferPtr1[i];
                    }
                    break;
            }
        }

    }
    catch (interprocess_exception &ex) {
        std::cout << "getSyncImages: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }

}

void DataProxy::getImage(int &ts, int &tms, int &width, int &height, int &nbChnls, unsigned char *imgBuffer, int imageFormat)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm(open_only , shared_memory_video0_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region(shm , read_write);

        //Get a pointer to this shared memory block.
        void *addr = region.get_address();
        ImageMessage *data = static_cast<ImageMessage *>(addr);

        boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock(data->imageMutex);
        ts = data->ts;
        tms = data->tms;
        width = data->width;
        height = data->height;
        nbChnls = data->nbChannels;

        //Read the data buffer.
        unsigned char *bufferPtr = data->items;

        switch (imageFormat) {
            case RGB :
                for (int i = 0 ; i < width * height * nbChnls ; ++i) {
                    imgBuffer[i] = bufferPtr[i];
                }
                break;
            case BGR :
                for (int i = 0 ; i < width * height ; ++i) {
                    imgBuffer[3 * i] = bufferPtr[3 * i + 2];
                    imgBuffer[3 * i + 1] = bufferPtr[3 * i + 1];
                    imgBuffer[3 * i + 2] = bufferPtr[3 * i];
                }
                break;
            case MATLABRGB :
                for (int i = 0 ; i < width ; ++i)
                    for (int j = 0 ; j < height ; ++j) {
                        imgBuffer[i * height + j] = bufferPtr[3 * j * width + 3 * i];
                        imgBuffer[width * height + i * height + j] = bufferPtr[3 * j * width + 3 * i + 1];
                        imgBuffer[2 * width * height + i * height + j] = bufferPtr[3 * j * width + 3 * i + 2];
                    }
                break;
            default :
                for (int i = 0 ; i < width * height * nbChnls ; ++i) {
                    imgBuffer[i] = bufferPtr[i];
                }
                break;
        }

    }
    catch (interprocess_exception &ex) {
        std::cout << "getImage: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::getSyncImage(int &ts, int &tms, int &width, int &height, int &nbChnls, unsigned char *imgBuffer, int imageFormat)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm(open_only , shared_memory_video0_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region(shm , read_write);

        //Get a pointer to this shared memory block.
        void *addr = region.get_address();
        ImageMessage *data = static_cast<ImageMessage *>(addr);

        boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock(data->imageMutex);
        data->imageCond.wait(lock);
        ts = data->ts;
        tms = data->tms;
        width = data->width;
        height = data->height;
        nbChnls = data->nbChannels;

        //Read the data buffer.
        unsigned char *bufferPtr = data->items;

        switch (imageFormat) {
            case RGB :
                for (int i = 0 ; i < width * height * nbChnls ; ++i) {
                    imgBuffer[i] = bufferPtr[i];
                }
                break;
            case BGR :
                for (int i = 0 ; i < width * height ; ++i) {
                    imgBuffer[3 * i] = bufferPtr[3 * i + 2];
                    imgBuffer[3 * i + 1] = bufferPtr[3 * i + 1];
                    imgBuffer[3 * i + 2] = bufferPtr[3 * i];
                }
                break;
            case MATLABRGB :
                for (int i = 0 ; i < width ; ++i)
                    for (int j = 0 ; j < height ; ++j) {
                        imgBuffer[i * height + j] = bufferPtr[3 * j * width + 3 * i];
                        imgBuffer[width * height + i * height + j] = bufferPtr[3 * j * width + 3 * i + 1];
                        imgBuffer[2 * width * height + i * height + j] = bufferPtr[3 * j * width + 3 * i + 2];
                    }
                break;
            default :
                for (int i = 0 ; i < width * height * nbChnls ; ++i) {
                    imgBuffer[i] = bufferPtr[i];
                }
                break;
        }

    }
    catch (interprocess_exception &ex) {
        std::cout << "getSyncImage: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::getImageTimeStamp(int &ts, int &tms)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm(open_only , shared_memory_video0_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region(shm , read_write);

        //Get a pointer to this shared memory block.
        void *addr = region.get_address();
        ImageMessage *data = static_cast<ImageMessage *>(addr);

        boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock(data->imageMutex);
        ts = data->ts;
        tms = data->tms;

    }
    catch (interprocess_exception &ex) {
        std::cout << "getImageTimeStamp: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::getSoundBuffer(int &ts, int &tms, int &nbSamples , int &nbChnls, int &freqHz, short *soundBuffer)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm(open_only , shared_memory_audio_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region(shm , read_write);

        //Get a pointer to this shared memory block.
        void *addr = region.get_address();
        AudioMessage *aMes = static_cast<AudioMessage *>(addr);
        {
            boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock(aMes->soundMutex);
            ts = aMes->ts;
            tms = aMes->tms;
            nbSamples = aMes->samples;
            nbChnls = aMes->channels;
            freqHz = aMes->fRate;

            //Read the data buffer.
            memcpy(soundBuffer, aMes->data, nbSamples * nbChnls * sizeof(short));
        }

    }
    catch (interprocess_exception &ex) {
        std::cout << "getSoundBuffer: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::getSyncSoundBuffer(int &ts, int &tms, int &nbSamples , int &nbChnls, int &freqHz, short *soundBuffer)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm(open_only , shared_memory_audio_name.c_str(), read_write);

        //Map the whole shared memory in this process
        mapped_region region(shm , read_write);

        //Get a pointer to this shared memory block.
        void *addr = region.get_address();
        AudioMessage *aMes = static_cast<AudioMessage *>(addr);

        {
            boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock(aMes->soundMutex);
            aMes->soundCond.wait(lock);
            ts = aMes->ts;
            tms = aMes->tms;
            nbSamples = aMes->samples;
            nbChnls = aMes->channels;
            freqHz = aMes->fRate;

            //Read the data buffer.
            memcpy(soundBuffer, aMes->data, nbSamples * nbChnls * sizeof(short));

        }
    }
    catch (interprocess_exception &ex) {
        std::cout << "getSoundBuffer: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}

void DataProxy::getSoundBufferTimeStamp(int &ts, int &tms)
{
    try {
        //Open the shared memory object.
        shared_memory_object shm(open_only , shared_memory_audio_name.c_str(), read_only);

        //Map the whole shared memory in this process
        mapped_region region(shm , read_only);

        void *addr = region.get_address();
        AudioMessage *aMes = static_cast<AudioMessage *>(addr);

        {
            boost::interprocess::sharable_lock<boost::interprocess::interprocess_upgradable_mutex> lock(aMes->soundMutex);
            //Read timestamp second.
            ts = aMes->ts;
            //Read timestamp microsecond.
            tms = aMes->tms;
        }
    }
    catch (interprocess_exception &ex) {
        std::cout << "getSoundBufferTimeStamp: " << ex.what() << std::endl;
        throw std::runtime_error(ex.what());
    }
}
