/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Quentin Perlorson, Fabien Badeig
 * Contacts :
 * perception-naolab@inria.fr
 * 
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
* @file dataproxies.h
* @author Pelorson Quentin, Fabien Badeig
* @date 2015 Feb 25
* @brief
**/

#ifndef DATAPROXIES_H
#define DATAPROXIES_H
#include <vector>
#include <string>

namespace DataProxy
{

enum ImgaeFormat {
    RGB,
    BGR,
    MATLABRGB
};

static int IMAGEWIDTH = 640;
static int IMAGEHEIGHT = 480;
static int IMAGENBCHANNELS = 3;
static int SOUNDFREQHZ = 48000;
static int SOUNDNBCHANNELS = 4;
static int SOUNDNBSAMPLES = 4096;

/** @brief Move the head of the robot according to the given angle. This is a blocking function.
  * Throw an std::runtime exception.
  * @param YawAngleValue yaw angle if absolute == 0 move the head of YawAngleValue radian from the current position. Absolute move LIMIT [-2.0, 2.0]
  * @param PitchAngleValue pitch angle if absolute == 0 move the head of YawAngleValue radian from the current position. Absolute move LIMIT [-0.6, 0.5]
  * @param absolute Is an absolute move or not.
  **/
void moveHead(float yawAngle, float pitchAngle, bool absolute, float &newAbsoluteYaw, float newAbsolutePitch);

/** @brief Move selected servo-motors of the robot according to the given angles. This is a blocking function.
  * Throw an std::runtime exception.
  * @param lJoints: list of motor ID
  * @param lAngles: list of absolute angles for each motor
  * @param lSpeeds: speed to make the move
  * @param nbElt: number of elements in each list
  * @param currentAngles: return the real motor position
  **/
void moveMotorTo(double *lJoints, double *lAngles, double *lSpeeds, int nbElt, float *currentAngles);

/** @brief Move Nao (stand up posture).
  * Throw an std::runtime exception.
  * @param x: distance on the x axis (in front of him)
  * @param y: distance on the y axis (on side)
  * @param theta: rotation of the robot
  **/
void moveToPoint(float &x, float &y, float &theta);


/** @brief Give angle info for each selected servo-motors.
  * Throw an std::runtime exception.
  * @param lJoints: list of motor ID
  * @param nbElt: number of elements in list lJoints
  * @param currentAngles: return the motor angles
  **/
void getMotorInfo(double *lJoints, int nbElt, float *currentAngles);

/** @brief Give angle info of all servo-motors.
  * Throw an std::runtime exception.
  * @param currentAngles: return all motor angles
  **/
void getAllMotorInfo( float *currentAngles );

/** @brief Get the robot face detector results.
  * Throw an std::runtime exception.
  * @param ts the timestamp. Second part.
  * @param tms the timestamp. MicroSecond part.
  * @param nbFaces the of faces detected.
  * @param facesAngles the image coordinate angles of detected faces stored in a vector:
  * alpha : yaw angle in the image referential. beta pitch angle in the image referential.
  * [alpah0 beta0 alpah1 beta1 alpah2 beta2....]
 **/
void getFaceDetectorData(int &ts, int &tms, int &nbFaces, std::vector<float> &facesAngles);
void getFaceDetectorSyncData(int &ts, int &tms, int &nbFaces, std::vector<float> &facesAngles);

/** @brief Get the current image at the RGB format.
  * Throw an std::runtime exception.
  * @param ts the timestamp. Second part.
  * @param tms the timestamp. MicroSecond part.
  * @param width the width of the image (pixel).
  * @param height the height of the image (pixel).
  * @param NbChnls the number of channels.
  * @param imgBuffer the data buffer, must be previously allocated. needed size :
  * RGB-VGA : 3*640*480.
  * @param imageFormat image format from the enum ImgaeFormat:
 **/
void getImage(int &ts, int &tms, int &width, int &height, int &nbChnls, unsigned char *imgBuffer, int imageFormat);
void getSyncImage(int &ts, int &tms, int &width, int &height, int &nbChnls, unsigned char *imgBuffer, int imageFormat);

/** @brief same job as the method getImage but this method grabs two images from NAO top/left and bottom/right cameras
 *
 **/
void getImages(int &ts0, int &tms0, int &width0, int &height0, int &nbChnls0, unsigned char *imgBuffer0, int &ts1, int &tms1, int &width1, int &height1, int &nbChnls1, unsigned char *imgBuffer1, int imageFormat);

void getSyncImages(int &ts0, int &tms0, int &width0, int &height0, int &nbChnls0, unsigned char *imgBuffer0, int &ts1, int &tms1, int &width1, int &height1, int &nbChnls1, unsigned char *imgBuffer1, int imageFormat);

/** @brief Get the current image's TimeStamp.
  * Throw an std::runtime exception.
  * @param ts the timestamp. Second part.
  * @param tms the timestamp. MicroSecond part.
 **/
void getImageTimeStamp(int &ts, int &tms);

/** @brief Get the current sound buffer.
  * Throw an std::runtime exception.
  * @param ts the timestamp. Second part.
  * @param tms the timestamp. MicroSecond part.
  * @param nbSamples the number of samples for each channel.
  * @param NbChnls the number of channels.
  * @param freqHz the Frequency in hz.
  * @param soundBuffer the data buffer, must be previously allocated. needed size : 16384.
  * Data stored :
  *    s1c1 s2c1 s3c1......
  *    s1c2 s2c2 s3c2......
  *    s1c3 s2c3 s3c3......
  *    s1c4 s2c4 s3c4......
  *
  * Micro
  * 1: RearLeft
  * 2: RearRight
  * 3: FrontLeft
  * 4: FrontRight
**/
void getSoundBuffer(int &ts, int &tms, int &nbSamples , int &nbChnls, int &freqHz, short *soundBuffer);
void getSyncSoundBuffer(int &ts, int &tms, int &nbSamples , int &nbChnls, int &freqHz, short *soundBuffer);

/** @brief Get the current sound buffer timestamp.
  * Throw an std::runtime exception.
  * @param ts the timestamp. Second part.
  * @param tms the timestamp. MicroSecond part.
**/
void getSoundBufferTimeStamp(int &ts, int &tms);

/** @brief Set the NAO posture.
 * Throw an std::runtime exception.
 * @param posture: desired posture.
 * @param speed: period to reach the posture.
**/
void setNaoPosture(std::string &posture, float &speed);

}

#endif // DATAPROXIES_H

