/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Quentin Perlorson, Fabien Badeig
 * Contacts :
 * perception-naolab@inria.fr
 *
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
 * @file naoaudiodriver.cpp
 * @author Pelorson Quentin, Fabien Badeig
 * @date 2015 Feb 15
 * @brief
 **/

#include "naoaudiodriver.h"

#include <boost/interprocess/shared_memory_object.hpp>
#include <boost/interprocess/mapped_region.hpp>

#define _DEBUG

using namespace boost::interprocess;

const int FREQ_RATE = 48000;
const int NB_CHANNELS = 4;
int MaxNofSamples = 16384;

/**
 * Constructor for NaoAudioDriver object
 * @param broker The parent broker
 * @param name The name of the module
 */
NaoAudioDriver::NaoAudioDriver(boost::shared_ptr<AL::ALBroker> broker):
    name("NaoAudioDevice")
    , connected(false)
{
    if (broker) {
        AL::ALModule::createModule<NaoAudioDevice>(broker, "NaoAudioDevice");
        soundModule  = boost::shared_ptr<AL::ALProxy>(new AL::ALProxy(broker, "NaoAudioDevice"));
        if (soundModule != NULL) {
            connected = true;
        }
        else {
            std::cout << "Error with NaoAudioDevice" << std::endl;
        }
    }
    pthread_mutex_init(&stopMutex, NULL);
    pthread_cond_init (&stopCond, NULL);

    shared_memory_object::remove(shared_memory_audio_name.c_str());
    //Create a shared memory object.
    shared_memory_object shm (create_only, shared_memory_audio_name.c_str(), read_write);

    //Set size
    shm.truncate(sizeof(AudioMessage));

    //Map the whole shared memory in this process
    mapped_region region(shm, read_write);

    void *addr = region.get_address();
    //Construct the shared structure in memory
    AudioMessage *aMes = new (addr) AudioMessage;
}

NaoAudioDriver::~NaoAudioDriver()
{
    soundModule.reset();
    pthread_mutex_destroy(&stopMutex);
    pthread_cond_destroy(&stopCond);
    shared_memory_object::remove(shared_memory_audio_name.c_str());
}

void NaoAudioDriver::startCapture()
{
    try {
        if (soundModule != NULL) {
            pthread_create(&fAudioThread, 0, xAudioThread, this);
            std::cout << "Audio thread created." << std::endl;
        }
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot subscribe to audio: " << error.what() << std::endl;
    }
}

void NaoAudioDriver::stopCapture()
{
    if (!soundModule && soundModule->call<bool>("isCapturing")) {
        pthread_mutex_lock(&stopMutex);
        pthread_cond_signal(&stopCond);
        pthread_mutex_unlock(&stopMutex);
        void *lReturn;
        pthread_join(fAudioThread, &lReturn);
        std::cout << "Audio thread closed." << std::endl;
    }
}

bool NaoAudioDriver::setBroker(boost::shared_ptr<AL::ALBroker> broker)
{
    if (soundModule != NULL) {
        if (soundModule->call<bool>("isCapturing")) {
            stopCapture();
        }
        soundModule.reset();
    }
    if (broker) {
        AL::ALModule::createModule<NaoAudioDevice>(broker, "NaoAudioDevice");
        soundModule  = boost::shared_ptr<AL::ALProxy>(new AL::ALProxy(broker, "NaoAudioDevice"));
    }
    else {
        soundModule = boost::shared_ptr<AL::ALProxy>();
    }
    if (soundModule != NULL) {
        connected = true;
    }
    else {
        connected = false;
    }
    return soundModule != NULL;
}

bool NaoAudioDriver::isConnected()
{
    return connected;
}

bool NaoAudioDriver::isCapturing()
{
    if (soundModule == NULL) {
        return false;
    }
    return soundModule->call<bool>("isCapturing");
}

void *NaoAudioDriver::xAudioThread(void *pArg)
{
    NaoAudioDriver *lThis = (NaoAudioDriver *)pArg;
    lThis->soundModule->callVoid("xStartAudio");
    pthread_mutex_lock(&lThis->stopMutex);
    pthread_cond_wait(&lThis->stopCond, &lThis->stopMutex);
    pthread_mutex_unlock(&lThis->stopMutex);
    lThis->soundModule->callVoid("xStopAudio");
    pthread_exit(NULL);
}

// NaoAudioDevice ============================================================

/**
 * Constructor for NaoAudioDevice object
 * @param broker The parent broker
 * @param name The name of the module
 */
NaoAudioDevice::NaoAudioDevice(
    boost::shared_ptr<AL::ALBroker> broker,
    const std::string &name)
    : AL::ALSoundExtractor(broker, "NaoAudioDevice")
    , fCapturingAudio(false)
{
    setModuleDescription("Nao Sound module.");

    functionName("xStartAudio", getName(), "Start audio capturing.");
    BIND_METHOD(NaoAudioDevice::xStartAudio);

    functionName("xStopAudio", getName(), "Stop audio capturing.");
    BIND_METHOD(NaoAudioDevice::xStopAudio);

    functionName("isCapturing", getName(), "Check if the sound is currently capturing");
    setReturn("bool", "return state of the capturing process");
    BIND_METHOD(NaoAudioDevice::isCapturing);

    ts = 0;
    tms = 0;
    tsF = -1;
    tmsF = -1;
}

NaoAudioDevice::~NaoAudioDevice()
{
}

void NaoAudioDevice::init()
{
    audioDevice->callVoid("setClientPreferences",
                          "NaoAudioDevice",                //Name of this module
                          FREQ_RATE,                    //16000Hz | 48000 Hz requested
                          (int)AL::ALLCHANNELS,     //4 Channels requested
                          0                         //Deinterleaving not requested
                         );
}

bool NaoAudioDevice::isCapturing()
{
    return fCapturingAudio;
}

void NaoAudioDevice::xStartAudio()
{

    try {
        audioDevice->callVoid("setClientPreferences",
                              "NaoAudioDevice",                //Name of this module
                              FREQ_RATE,                    //16000Hz | 48000 Hz requested
                              (int)AL::ALLCHANNELS,     //4 Channels requested
                              0                         //Deinterleaving not requested
                             );

        startDetection();
        std::cout << "STREAMING STARTED." << std::endl;
        fCapturingAudio = true;
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot subscribe audio: " << error.what() << std::endl;
        fCapturingAudio = false;
    }
}

void NaoAudioDevice::xStopAudio()
{
    try {
        stopDetection();
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot unsubscribe audio." << error.what() << std::endl;
    }
    fCapturingAudio = false;
}

void NaoAudioDevice::process(const int &pNbrChannels,
                             const int &pNbrSamples,
                             const AL_SOUND_FORMAT *pDataInterleaved,
                             const AL::ALValue &pTimeStamp)
{
    //Micro
    // 0: RearLeft
    // 1: RearRight
    // 2: FrontLeft
    // 3: FrontRight

    //Cast timestamp values to int
    //Timestamp second.
    ts = (int)pTimeStamp[0];
    //Timestamp millisecond.
    tms = (int)pTimeStamp[1];

    if (tsF == -1 && tmsF == -1) {
        tsF = ts;
        tmsF = tms;
    }

    //Open shared memory object.
    shared_memory_object shm (open_only, shared_memory_audio_name.c_str(), read_write);

    //Map the whole shared memory in this process
    mapped_region region(shm, read_write);

    void *addr = region.get_address();
    //Construct the shared structure in memory
    AudioMessage *aMes = static_cast<AudioMessage *>(addr);

    {
        boost::interprocess::scoped_lock<boost::interprocess::interprocess_upgradable_mutex> lock(aMes->soundMutex);
        aMes->ts = ts;
        aMes->tms = tms;
        aMes->samples = pNbrSamples;
        aMes->channels = pNbrChannels;
        aMes->fRate = FREQ_RATE;

        memcpy(aMes->data, pDataInterleaved, pNbrSamples * pNbrChannels * sizeof(short));
        aMes->soundCond.notify_all();
    }
    // debug display to keep
#ifdef _DEBUG
    std::cout << ts << "," << tms << " : " << pNbrSamples << " samples, " << pNbrChannels << " / " << NB_CHANNELS << " channels, " << FREQ_RATE << "Hz" << std::endl;
#endif

}
