/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Quentin Perlorson, Fabien Badeig
 * Contacts :
 * perception-naolab@inria.fr
 * 
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
* @file naoaudiodriver.h
* @author Pelorson Quentin, Fabien Badeig
* @date 2015 Feb 15
* @brief
**/

#ifndef NAOAUDIODRIVER_H
#define NAOAUDIODRIVER_H

#include <string>
#include <pthread.h>
#include "commondefinitions.h"

#ifndef Q_MOC_RUN
#include <alcommon/almodule.h>
#include <alcommon/alproxy.h>
#include <alaudio/alsoundextractor.h>
#include <alproxies/altexttospeechproxy.h>
#endif // Q_MOC_RUN

namespace AL
{
class ALBroker;
}
class NaoAudioDevice;

/**
 * Remote module for audio capture.
 */
class NaoAudioDriver
{
public:
    /**
     * Default Constructor for modules.
     * @param broker the broker to which the module should register.
     */
    NaoAudioDriver(boost::shared_ptr<AL::ALBroker> broker);

    /**  Destructor. **/
    virtual ~NaoAudioDriver();

    /**
     * Launch the record process.
     * Write on the disk the received sounds in the shared memory .
    **/
    void startCapture();

    /** Stop capture. **/
    void stopCapture();

    /**
    * Set the broker to use.
    * @param broker the broker to which the module should register.
    * @return TRUE if connection via broker is successful, FALSE otherwise.
    **/
    bool setBroker(boost::shared_ptr<AL::ALBroker> broker);

    /**
     * Return the state of the connection to the video device on nao.
     * @return TRUE if connected to nao, FALSE otherwise.
    **/
    bool isConnected();

    /**
     * Return the state of the capturing process.
     * @return TRUE if the capturing process is runnings, FALSE otherwise.
    **/
    bool isCapturing();

private:

    /** Audio acquisition thread.
     * @param pArg user data, here the module instance.
     * @return Thread return value
    **/
    static void *xAudioThread(void *pArg);

    /** Audio thread object. **/
    pthread_t fAudioThread;

    /** Proxy to audio device. **/
    //NaoAudioDevice *naoAudioDevice;
    boost::shared_ptr<AL::ALProxy> soundModule;

    /** Module name. **/
    std::string name;

    /** Stop record thread mutex.**/
    pthread_mutex_t stopMutex;

    /** Stop record thread condition. **/
    pthread_cond_t stopCond;

    /** Are we conected ?**/
    bool connected;
};

class NaoAudioDevice : public AL::ALSoundExtractor
{
public:
    /**
     * Default Constructor for modules.
     * @param broker the broker to which the module should register.
     * @param name the boadcasted name of the module.
     */
    NaoAudioDevice(boost::shared_ptr<AL::ALBroker> broker, const std::string &name);

    /** Destructor. **/
    virtual ~NaoAudioDevice();

    /**
     * Are we currently capturing ?
     * @return whether we started capturing.
     **/
    bool isCapturing();

    void init();

    /**
     * Remote callback from ALAudioDevice providing sound buffers.
     * @param pNbOfInputChannels number of audio channels provided here.
     * @param pNbrSamples length of the audio buffer in samples.
     * @param pDataInterleaved raw interleaved audio buffer.
     * @param pTimeStamp The time stamp of the audio buffer
     */
    void process(const int &pNbrChannels,
                 const int &pNbrSamples,
                 const AL_SOUND_FORMAT *pDataInterleaved,
                 const AL::ALValue &pTimeStamp);

    /** Start audio capture. **/
    void xStartAudio();

    /** Stop audio capture. **/
    void xStopAudio();

    /** Are we currently capturing audio ? **/
    bool fCapturingAudio;

    int ts;
    int tms;
    int tsF;
    int tmsF;

};
#endif  // NAOAUDIODRIVER_H

