/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Quentin Perlorson
 * Contacts :
 * perception-naolab@inria.fr
 *
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
 * @file naofacedetectordriver.cpp
 * @author Pelorson Quentin, Fabien Badeig
 * @date 2015 Feb 26
 * @brief
 **/

#include "naofacedetectordriver.h"

using namespace boost::interprocess;


// NaoFaceDetectorDriver Class ==========================================================
/**
 * Constructor for NaoFaceDetectorDriver object
 * @param broker The parent broker
 */
NaoFaceDetectorDriver::NaoFaceDetectorDriver(boost::shared_ptr<AL::ALBroker> broker):
    fMemProxy(0),
    fConnected(false),
    fSharedMemoryStructPtr(0)
{
    if (broker) {
        fMemProxy = new AL::ALMemoryProxy(broker);
        if (fMemProxy) {
            fMemProxy->subscribeToEvent("FaceDetected", "", "");
            fConnected = 1;
        }
    }
}

NaoFaceDetectorDriver::~NaoFaceDetectorDriver()
{
    if (fMemProxy) {
        fMemProxy->unsubscribeToEvent("FaceDetected", "");
        delete fMemProxy;
    }
    shared_memory_object::remove(shared_memory_facedetector_name.c_str());
    std::cout << "NaoFaceDetectorDriver destroyed." << std::endl;
}

bool NaoFaceDetectorDriver::setBroker(boost::shared_ptr<AL::ALBroker> broker)
{
    if (fMemProxy) {
        delete fMemProxy;
    }
    if (broker) {
        fMemProxy = new AL::ALMemoryProxy(broker);
    }
    else {
        fMemProxy = 0;
    }
    if (fMemProxy) {
        fConnected = true;
    }
    else {
        fConnected = false;
    }
    return fMemProxy != 0;
}

bool NaoFaceDetectorDriver::isConnected()
{
    return fConnected;
}

void NaoFaceDetectorDriver::getFaceDetectionData()
{
    AL::ALValue data = fMemProxy->getData("FaceDetected");
    if (!data.getSize()) {
        AL::ALValue timeStamp = fMemProxy->getTimestamp("FaceDetected");
        fSharedMemoryStructPtr->nbFaces = 0;
        if (fSharedMemoryStructPtr->ts == (int)timeStamp[1] && fSharedMemoryStructPtr->tms == (int)timeStamp[2]) {
            return;
        }
        fSharedMemoryStructPtr->ts = timeStamp[1];
        fSharedMemoryStructPtr->tms = timeStamp[2];
        return;
    }

    AL::ALValue timeStamp = data[0];
    if (fSharedMemoryStructPtr->ts == (int)timeStamp[1] && fSharedMemoryStructPtr->tms == (int)timeStamp[2]) {
        return;
    }

    AL::ALValue faceInfoArray = data[1];
    int nbFacesDetected = faceInfoArray.getSize() - 1;
    fSharedMemoryStructPtr->nbFaces = nbFacesDetected;
    fSharedMemoryStructPtr->ts = timeStamp[0];
    fSharedMemoryStructPtr->tms = timeStamp[1];
    for (int i = 0 ; i < nbFacesDetected ; i++) {
        AL::ALValue currentFace = faceInfoArray[i];
        AL::ALValue currentFaceInfo = currentFace[0];
        if (i < 50) {
            fSharedMemoryStructPtr->items[2 * i] = currentFaceInfo[1];
            fSharedMemoryStructPtr->items[2 * i + 1] = currentFaceInfo[2];
        }
    }
    std::cout << timeStamp[0] << "," << timeStamp[1]  << " : " << nbFacesDetected << " faces" << std::endl;
}

void NaoFaceDetectorDriver::waitLoop()
{
    try {
        //Remove a shared memory object if it exists.
        shared_memory_object::remove(shared_memory_facedetector_name.c_str());

        //Create a shared memory object.
        shared_memory_object shm (create_only, shared_memory_facedetector_name.c_str(), read_write);

        //Set size
        shm.truncate(sizeof(FaceDetectorMessage));

        //Map the whole shared memory in this process
        mapped_region region(shm, read_write);

        //Get the pointer to the shared memory block;
        void *addr = region.get_address();

        fSharedMemoryStructPtr = new (addr) FaceDetectorMessage;

        std::cout << "STREAMING STARTED." << std::endl;

        while (1) {
            {

                boost::interprocess::scoped_lock<boost::interprocess::interprocess_upgradable_mutex> lock(fSharedMemoryStructPtr->fdetectMutex);

                getFaceDetectionData();
            }
            boost::this_thread::sleep( boost::posix_time::milliseconds(20) );
            fSharedMemoryStructPtr->fdetectCond.notify_all();
        }
    }
    catch (const std::exception &error) {
        std::cerr << "NaoFaceDetectorDriver stopped reason: " << error.what() << std::endl;
    }
}
