/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Quentin Perlorson, Fabien Badeig
 * Contacts :
 * perception-naolab@inria.fr
 * 
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
* @file naoheadmotiondriver.cpp
* @author Pelorson Quentin, Fabien Badeig
* @date 2015 Feb 25
* @brief
**/

#include "naoheadmotiondriver.h"

#include "commondefinitions.h"

using namespace boost::interprocess;
float yawMin = -1.5f;
float yawMax = 1.5f;
float pitchMin = -0.6f;
float pitchMax = 0.5f;


// NaoMotionDriver Class ==========================================================
/**
 * Constructor for NaoMotionDriver object
 * @param broker The parent broker
 */
NaoHeadMotionDriver::NaoHeadMotionDriver(boost::shared_ptr<AL::ALBroker> broker):
    fMotionCommandHandlerThread(),
    fJoinsNames(AL::ALValue::array("HeadYaw", "HeadPitch")),
    fAngleLists(AL::ALValue::array(0.0f, 0.0f)),
    fTimeLists(AL::ALValue::array(0.4f, 0.4f)),
    fIsAbsolute(true),
    fThreadRunning(false)
{
    if (broker) {
        fALMotionProxy = new AL::ALMotionProxy(broker);
    }
}

NaoHeadMotionDriver::~NaoHeadMotionDriver()
{
    delete fALMotionProxy;
    shared_memory_object::remove(shared_memory_head_motion_name.c_str());
}

bool NaoHeadMotionDriver::setBroker(boost::shared_ptr<AL::ALBroker> broker)
{
    if (fALMotionProxy) {
        if (isWakeUp()) {
            rest();
        }
        delete fALMotionProxy;
    }
    if (broker) {
        fALMotionProxy = new AL::ALMotionProxy(broker);
    }
    else {
        fALMotionProxy = 0;
    }
    return fALMotionProxy != 0;
}

void NaoHeadMotionDriver::wakeUp()
{
    try {
        if (!fALMotionProxy) {
            throw std::runtime_error("Need to create a valid broker before.");
            return;
        }

        if (isWakeUp()) {
            throw std::runtime_error("Robot already wake up.");
        }
        else {
            fALMotionProxy->wakeUp();
            if (isWakeUp()) {
                fALMotionProxy->angleInterpolation(fJoinsNames, fAngleLists, fTimeLists, fIsAbsolute);
                std::cout << "NAO IS WAKE UP AND READY." << std::endl;
                xStartHeadMotionCommandHandler();
            }
        }
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot wake up : " << error.what() << std::endl;
    }
}

void NaoHeadMotionDriver::rest()
{
    if (isWakeUp()) {
        xStopHeadMotionCommandHandler();
        fALMotionProxy->rest();
        std::cout << "Robot Motion stopped." << std::endl;
    }
}

bool NaoHeadMotionDriver::isWakeUp()
{
    if (fALMotionProxy) {
        return fALMotionProxy->robotIsWakeUp();
    }
    return 0;
}

void NaoHeadMotionDriver::moveHeadTo(float yaw, float pitch)
{
    std::vector<float> currentAngles = fALMotionProxy->getAngles(fJoinsNames, true);
    float newYaw = currentAngles[0] + yaw;
    float newPitch = currentAngles[1] + pitch;

    newYaw = (newYaw > yawMax) ? yawMax : newYaw;
    newYaw = (newYaw < yawMin) ? yawMin : newYaw;
    newPitch = (newPitch > pitchMax) ? pitchMax : newPitch;
    newPitch = (newPitch < pitchMin) ? pitchMin : newPitch;

    fAngleLists[0] = newYaw;
    fAngleLists[1] = newPitch;
    fALMotionProxy->angleInterpolation(fJoinsNames, fAngleLists, fTimeLists, fIsAbsolute);
    //std::cout << fALMotionProxy->getSummary() << std::endl;
}

void NaoHeadMotionDriver::absoluteMoveHeadTo(float yaw, float pitch)
{
    float newYaw = yaw;
    float newPitch = pitch;

    newYaw = (newYaw > yawMax) ? yawMax : newYaw;
    newYaw = (newYaw < yawMin) ? yawMin : newYaw;
    newPitch = (newPitch > pitchMax) ? pitchMax : newPitch;
    newPitch = (newPitch < pitchMin) ? pitchMin : newPitch;

    fAngleLists[0] = newYaw;
    fAngleLists[1] = newPitch;
    fALMotionProxy->angleInterpolation(fJoinsNames, fAngleLists, fTimeLists, true);
}

std::vector<float> NaoHeadMotionDriver::getHeadPosition()
{
    std::vector<float> currentPosition = fALMotionProxy->getAngles(fJoinsNames, true);
    return currentPosition;
}


void NaoHeadMotionDriver::xStartHeadMotionCommandHandler()
{
    try {
        pthread_create(&fMotionCommandHandlerThread, 0, xHeadMotionCommandHandlerThread, this);
        fThreadRunning = true;
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot launch Motion Command Handler Thread: " << error.what() << std::endl;
    }
}

void NaoHeadMotionDriver::xStopHeadMotionCommandHandler()
{
    if (fThreadRunning) {
        pthread_cancel(fMotionCommandHandlerThread);
        fThreadRunning = false;
        std::cout << "Motion Command Handler Thread closed." << std::endl;
    }
}

// Thread =====================================================================
//Frame grabbing function.
void *NaoHeadMotionDriver::xHeadMotionCommandHandlerThread(void *pArg)
{
    //The calling object.
    NaoHeadMotionDriver *lThis = (NaoHeadMotionDriver *)pArg;

    shared_memory_object::remove(shared_memory_head_motion_name.c_str());
    //Create a shared memory object.
    shared_memory_object shm (open_or_create, shared_memory_head_motion_name.c_str(), read_write);

    //Set size
    shm.truncate(sizeof(HeadMotionMessage));

    //Map the whole shared memory in this process
    mapped_region region(shm, read_write);

    void *addr = region.get_address();
    //Construct the shared structure in memory
    HeadMotionMessage *data = new (addr) HeadMotionMessage;
    data->absoluteMove = false;
    data->items[0] = 0.0f;
    data->items[1] = 0.0f;

    try {
        while (1) {
            scoped_lock<interprocess_mutex> lock(data->newCommandMutex);
            data->newCommandCond.wait(lock);
            if (data->absoluteMove) {
                lThis->absoluteMoveHeadTo(data->items[0], data->items[1]);
            }
            else {
                lThis->moveHeadTo(data->items[0], data->items[1]);
            }
            std::cout << "Head command control = " << data->items[0] << " -- " << data->items[1] << std::endl;

            std::vector<float> currentPosition = lThis->getHeadPosition();
            data->items[2] = currentPosition[0];
            data->items[3] = currentPosition[1];
            data->motionFinishedCond.notify_one();
        }
    }
    catch (const std::exception &error) {
        std::cerr << "NaoHeadMotionDriver stopped reason: " << error.what() << std::endl;
    }
    pthread_exit(NULL);
}
