/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Quentin Perlorson, Fabien Badeig
 * Contacts :
 * perception-naolab@inria.fr
 *
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
 * @file naoheadmotiondriver.h
 * @author Pelorson Quentin, Fabien Badeig
 * @date 2015 Feb 25
 * @brief
 **/

#ifndef NAOHEADMOTIONDRIVER_H
#define NAOHEADMOTIONDRIVER_H

#include <string>
#include <pthread.h>
#include <boost/interprocess/sync/scoped_lock.hpp>
#include <boost/interprocess/shared_memory_object.hpp>
#include <boost/interprocess/mapped_region.hpp>
#include <boost/interprocess/sync/interprocess_mutex.hpp>
#include <boost/interprocess/sync/interprocess_condition.hpp>

#ifndef Q_MOC_RUN
#include <alcommon/almodule.h>
#include <alcommon/alproxy.h>
#include <alproxies/almotionproxy.h>
#endif // Q_MOC_RUN

namespace AL
{
    class ALBroker;
}

/**
 * Remote module for synchronized audio and video capture.
 */
class NaoHeadMotionDriver
{
 public:
    /**
     * Default Constructor for modules.
     * @param broker the broker to which the module should register.
     */
    NaoHeadMotionDriver(boost::shared_ptr<AL::ALBroker> broker);

    /** Destructor.**/
    ~NaoHeadMotionDriver();

    /**
     * Set the broker to use.
     * @param broker the broker to which the module should register.
     * @return TRUE if connection via broker is successful, FALSE otherwise.
     **/
    bool setBroker(boost::shared_ptr<AL::ALBroker> broker);

    /**
     * Start capture.
     * @param camResolution the video resolution see CamResolution.
     */
    void wakeUp();

    /** Stop capture.**/
    void rest();

    /**
     * Return the state of the connection to the video device on nao.
     * @return TRUE if connected to nao, FALSE otherwise.
     **/
    bool isWakeUp();

    void moveHeadTo(float yaw, float pitch);

    void absoluteMoveHeadTo(float yaw, float pitch);

    std::vector<float> getHeadPosition();

 private:
    /** Start Nao Motion. **/
    void xStartHeadMotionCommandHandler();

    /** Stop Nao Motion. **/
    void xStopHeadMotionCommandHandler();

    /** Launch motion control loop.
     * @param pArg user data, here the module instance.
     * @return Thread return value
     **/
    static void *xHeadMotionCommandHandlerThread(void *pArg);

    /** Proxy to Motion device. **/
    AL::ALMotionProxy *fALMotionProxy;

    /** Video thread object. **/
    pthread_t fMotionCommandHandlerThread;
    AL::ALValue fJoinsNames;
    AL::ALValue fAngleLists;
    AL::ALValue fTimeLists;
    bool fIsAbsolute;
    bool fThreadRunning;

};

#endif  // NAOHEADMOTIONDRIVER_H
