/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Fabien Badeig
 * Contacts :
 * perception-naolab@inria.fr
 *
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
* @file naomotiondriver.cpp
* @author Fabien Badeig
* @brief Driver to manage remotely Nao servo-motors
**/

#include "naomotiondriver.h"

#include "commondefinitions.h"

using namespace boost::interprocess;

// NaoMotionDriver Class ==========================================================
/**
 * Constructor for NaoMotionDriver object
 * @param broker The parent broker
 */
NaoMotionDriver::NaoMotionDriver(boost::shared_ptr<AL::ALBroker> broker):
    fMotionCommandHandlerThread(),
    fIsAbsolute(true),
    fMotionThreadRunning(false),
    fMoveThreadRunning(false),
    fPostureThreadRunning(false)
{
    if (broker) {
        setMotionProxy(broker);
        setPostureProxy(broker);
    }
    nbJoints = 0;
    fJoinsNamesAll.arraySetSize(26);
    for (int j = 0; j < 26; j++) {
        fJoinsNamesAll[j] = NAO_control_table[j].name;
    }
}

NaoMotionDriver::~NaoMotionDriver()
{
    postureProxy->goToPosture("Crouch", 0.5f);
    motionProxy->rest();
    delete postureProxy;
    delete motionProxy;
    delete[] jointID;
    shared_memory_object::remove(shared_memory_motion_control_name.c_str());
    shared_memory_object::remove(shared_memory_motion_contentg_name.c_str());
    shared_memory_object::remove(shared_memory_motion_contentp_name.c_str());
}

bool NaoMotionDriver::setMotionProxy(boost::shared_ptr<AL::ALBroker> broker)
{
    if (motionProxy = 0) {
        rest();
        delete motionProxy;
    }
    if (broker) {
        motionProxy = new AL::ALMotionProxy(broker);
    }
    else {
        motionProxy = 0;
    }
    return motionProxy != 0;
}

bool NaoMotionDriver::setPostureProxy(boost::shared_ptr<AL::ALBroker> broker)
{
    if (broker) {
        postureProxy = new AL::ALRobotPostureProxy(broker);
    }
    else {
        postureProxy = 0;
    }
    return postureProxy != 0;

}

void NaoMotionDriver::setPosture(std::string name, float speed)
{
    if (postureProxy != 0) {
        postureProxy->goToPosture(name, speed);
    }
}

std::string NaoMotionDriver::getPosture()
{
    std::string tempS("");
    if (postureProxy != 0) {
        tempS = postureProxy->getPosture();
    }
    return tempS;
}

void NaoMotionDriver::setJointNames(int *joints, int nbJ)
{
    if (nbJoints != 0 ) {
        delete[] jointID;
    }
    nbJoints = nbJ;
    jointID = new int[nbJoints];
    for (int i = 0; i < nbJoints; i++) {
        jointID[i] = joints[i];
        std::cerr << jointID[i] << " ** ";
    }
    std::cerr << std::endl;
    BuildJointsName();
}

void NaoMotionDriver::BuildJointsName()
{
    if ( nbJoints == 0 ) {
        return;
    }
    fJoinsNames.arraySetSize(nbJoints);
    fAngleLists.arraySetSize(nbJoints);
    fTimeLists.arraySetSize(nbJoints);
    for (int i = 0; i < nbJoints; i++) {
        for (int j = 0; j < 26; j++) {
            if ( NAO_control_table[j].ID == jointID[i] ) {
                fJoinsNames[i] = NAO_control_table[j].name;
                break;
            }
        }
    }
}

void NaoMotionDriver::setStiffness( bool lockJoints)
{
    if ( lockJoints == true ) {
        AL::ALValue fstiffnessList;
        AL::ALValue fstiffnessTime;
        fstiffnessList.arraySetSize(nbJoints);
        fstiffnessTime.arraySetSize(nbJoints);
        for (int i = 0; i < fJoinsNames.getSize(); i++) {
            fstiffnessList[i] = AL::ALValue::array(1.0f);
            fstiffnessTime[i] = AL::ALValue::array(0.2f);
        }
        motionProxy->setCollisionProtectionEnabled("Arms", true);
        motionProxy->stiffnessInterpolation(fJoinsNames, fstiffnessList, fstiffnessTime);
        qi::os::sleep(0.2f);
    }
    else {
        AL::ALValue fstiffnessList;
        AL::ALValue fstiffnessTime;
        fstiffnessList.arraySetSize(fJoinsNames.getSize());
        fstiffnessTime.arraySetSize(fJoinsNames.getSize());
        for (int i = 0; i < fJoinsNames.getSize(); i++) {
            fstiffnessList[i] = AL::ALValue::array(0.0f);
            fstiffnessTime[i] = AL::ALValue::array(0.2f);
        }
        motionProxy->stiffnessInterpolation(fJoinsNames, fstiffnessList, fstiffnessTime);
        qi::os::sleep(0.2f);
    }
}

bool NaoMotionDriver::IsRest()
{
    if (motionProxy) {
        std::vector<float> tempV = motionProxy->getStiffnesses(fJoinsNamesAll);
        for (int i = 0; i < tempV.size(); i++) {
            if (tempV[i] != 0.0f) {
                return false;
            }
        }
        return true;
    }
    return false;
}

void NaoMotionDriver::wakeUp()
{
    try {
        if (!motionProxy) {
            throw std::runtime_error("Need to create a valid proxy before.");
            return;
        }

        setStiffness(true);
        for (int i = 0; i < fJoinsNames.getSize(); i++) {
            fAngleLists[i] = AL::ALValue::array(0.0f);
            fTimeLists[i] = AL::ALValue::array(1.0f);
        }
        motionProxy->angleInterpolationBezier(fJoinsNames, fTimeLists, fAngleLists);
        std::cout << "NAO IS WAKE UP AND READY." << std::endl;
        xStartMotionCommandHandler();
        xStartPostureHandler();
        xStartMoveCommandHandler();
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot wake up : " << error.what() << std::endl;
    }
}

void NaoMotionDriver::rest()
{
    if (isWakeUp()) {
        xStopMotionCommandHandler();
        xStopPostureHandler();
        xStopMoveCommandHandler();
        setStiffness(false);
        std::cout << "Robot Motion stopped." << std::endl;
    }
}

bool NaoMotionDriver::isWakeUp()
{
    if (motionProxy) {
        std::vector<float> temp = motionProxy->getStiffnesses(fJoinsNames);

        for (int i = 0; i < temp.size(); i++ ) {
            if (temp[i] != 1.0f) {
                return false;
            }
        }
        return true;
    }
    return false;
}


std::vector<float> NaoMotionDriver::getAllJointInfos()
{
    return motionProxy->getAngles(fJoinsNamesAll, true);
}

std::vector<float> NaoMotionDriver::getJointInfos()
{
    return motionProxy->getAngles(fJoinsNames, true);
}


std::string NaoMotionDriver::getAllInfos()
{
    std::string tmp("");
    if (motionProxy) {
        tmp = motionProxy->getSummary();
    }
    return tmp;
}

void NaoMotionDriver::moveTo(float x, float y, float theta)
{
    if (motionProxy) {
        motionProxy->moveTo(x, y, theta);
    }
}

void NaoMotionDriver::moveTo(NaoPoint toPoint)
{
    if (motionProxy) {
        motionProxy->moveTo(toPoint.x, toPoint.y, toPoint.theta);
    }
}

void NaoMotionDriver::moveTo(std::vector<NaoPoint> &toPoints)
{
    if (motionProxy) {
        AL::ALValue arrayNP;
        arrayNP.arraySetSize(toPoints.size());
        for (int i = 0; i < toPoints.size(); i++) {
            arrayNP[i] = AL::ALValue::array(toPoints[i].x, toPoints[i].y, toPoints[i].theta);
        }
        motionProxy->moveTo(arrayNP);
    }
}

bool NaoMotionDriver::moveTo(float *jointAngles, float *jointSpeeds, int nbE)
{
    if ( nbE != fJoinsNames.getSize()) {
        return false;
    }

    for (int i = 0; i < fJoinsNames.getSize(); i++) {
        int t = 0;
        for (; t < 26; t++) {
            if (NAO_control_table[t].ID == jointID[i]) {
                break;
            }
        }
        if (t >= 26) {
            return false;
        }
        float tempA;
        jointAngles[jointID[i]] > NAO_control_table[t].max_collision_angle ? tempA = NAO_control_table[t].max_collision_angle : tempA = jointAngles[jointID[i]];
        jointAngles[jointID[i]] < NAO_control_table[t].min_collision_angle ? tempA = NAO_control_table[t].min_collision_angle : tempA = jointAngles[jointID[i]];
        fAngleLists[i] = AL::ALValue::array(tempA);
        fTimeLists[i] = AL::ALValue::array(jointSpeeds[jointID[i]]);
    }

    motionProxy->angleInterpolationBezier(fJoinsNames, fTimeLists, fAngleLists);
    std::cerr << "NAO IS MOVING ---> " << motionProxy->getAngles(fJoinsNames, true) << " <---" << std::endl;
    return true;
}

bool NaoMotionDriver::checkJoints(bool *maskID, int nbE)
{
    if ( nbE != fJoinsNames.getSize()) {
        return false;
    }

    for (int i = 0; i < fJoinsNames.getSize(); i++) {
        if ( maskID[jointID[i]] == false ) {
            return false;
        }

    }
    return true;
}

void NaoMotionDriver::xStartMotionCommandHandler()
{
    try {
        pthread_create(&fMotionCommandHandlerThread, 0, xMotionCommandHandlerThread, this);
        fMotionThreadRunning = true;
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot launch Motion Command Handler Thread: " << error.what() << std::endl;
    }
}

void NaoMotionDriver::xStopMotionCommandHandler()
{
    if (fMotionThreadRunning) {
        pthread_cancel(fMotionCommandHandlerThread);
        fMotionThreadRunning = false;
        std::cout << "Motion Command Handler Thread closed." << std::endl;
    }
}

void NaoMotionDriver::xStartMoveCommandHandler()
{
    try {
        pthread_create(&fMoveCommandHandlerThread, 0, xMoveCommandHandlerThread, this);
        fMoveThreadRunning = true;
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot launch Motion Command Handler Thread: " << error.what() << std::endl;
    }
}

void NaoMotionDriver::xStopMoveCommandHandler()
{
    if (fMoveThreadRunning) {
        pthread_cancel(fMoveCommandHandlerThread);
        fMoveThreadRunning = false;
        std::cout << "Motion Command Handler Thread closed." << std::endl;
    }
}

void NaoMotionDriver::xStartPostureHandler()
{
    try {
        pthread_create(&fPostureHandlerThread, 0, xPostureHandlerThread, this);
        fPostureThreadRunning = true;
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot launch posture Command Handler Thread: " << error.what() << std::endl;
    }
}

void NaoMotionDriver::xStopPostureHandler()
{
    if (fPostureThreadRunning) {
        pthread_cancel(fPostureHandlerThread);
        fPostureThreadRunning = false;
        std::cout << "Posture Command Handler Thread closed." << std::endl;
    }
}

// Thread =====================================================================
//Frame grabbing function.
void *NaoMotionDriver::xMotionCommandHandlerThread(void *pArg)
{
    //The calling object.
    NaoMotionDriver *lThis = (NaoMotionDriver *)pArg;

    shared_memory_object::remove(shared_memory_motion_control_name.c_str());
    //Create a shared memory object.
    shared_memory_object shm (open_or_create, shared_memory_motion_control_name.c_str(), read_write);

    //Set size
    shm.truncate(sizeof(MotionMessageControl));

    //Map the whole shared memory in this process
    mapped_region region(shm, read_write);

    void *addr = region.get_address();
    //Construct the shared structure in memory
    MotionMessageControl *data = new (addr) MotionMessageControl;

    shared_memory_object::remove(shared_memory_motion_contentg_name.c_str());
    //Create a shared memory object.
    shared_memory_object shmG (open_or_create, shared_memory_motion_contentg_name.c_str(), read_write);

    //Set size
    shmG.truncate(sizeof(MotionMessageContentG));

    //Map the whole shared memory in this process
    mapped_region regionG(shmG, read_write);

    void *addrG = regionG.get_address();
    //Construct the shared structure in memory
    MotionMessageContentG *contentG = new (addrG) MotionMessageContentG;

    shared_memory_object::remove(shared_memory_motion_contentp_name.c_str());
    //Create a shared memory object.
    shared_memory_object shmP (open_or_create, shared_memory_motion_contentp_name.c_str(), read_write);

    //Set size
    shmP.truncate(sizeof(MotionMessageContentP));

    //Map the whole shared memory in this process
    mapped_region regionP(shmP, read_write);

    void *addrP = regionP.get_address();
    //Construct the shared structure in memory
    MotionMessageContentP *contentP = new (addrP) MotionMessageContentP;

    try {
        while (1) {
            scoped_lock<interprocess_mutex> lock(data->newCommandMutex);
            data->newCommandCond.wait(lock);

            std::cerr << "Event Received!" << std::endl;

            if (strcmp(data->type, "get") == 0) {
                if ( data->jointArrayLength == -1) {
                    std::cerr << lThis->getAllInfos() << std::endl;
                    std::vector<float> currentPosition = lThis->getAllJointInfos();

                    for (int i = 0; i < 26; i++) {
                        contentG->currentPositions[i] = currentPosition[i];
                        contentG->mask[i] = true;
                    }
                }
                else {
                    std::vector<float> currentPosition = lThis->getJointInfos();

                    for (int i = 0; i < 26; i++) {
                        int j = 0;
                        for (; j < lThis->nbJoints; j++) {
                            if (i == lThis->jointID[j]) {
                                contentG->currentPositions[i] = currentPosition[j];
                                contentG->mask[i] = true;
                                break;
                            }
                        }
                        if ( j >= lThis->nbJoints ) {
                            contentG->mask[i] = false;
                        }
                    }
                }
            }
            else if (strcmp(data->type, "put") == 0) {

                if (lThis->checkJoints(contentP->mask, data->jointArrayLength) == false) {
                    lThis->setStiffness(false);
                    int *tempJ = new int[data->jointArrayLength];
                    int nbJE = 0;
                    for (int ji = 0; ji < 26; ji++) {
                        if ( contentP->mask[ji] == true ) {
                            tempJ[nbJE] = ji;
                            nbJE++;
                        }
                    }
                    lThis->setJointNames(tempJ, data->jointArrayLength);
                    lThis->setStiffness(true);
                    delete[] tempJ;
                }
                std::string collisionState = lThis->motionProxy->isCollision("Arms");

                std::cout <<  " Arms collision state: " << collisionState << std::endl;

                lThis->moveTo(contentP->angleLists, contentP->angleSpeeds, data->jointArrayLength);


                std::vector<float> currentPosition = lThis->getJointInfos();

                for (int i = 0; i < 26; i++) {
                    int j = 0;
                    for (; j < lThis->nbJoints; j++) {
                        if (i == lThis->jointID[j]) {
                            contentG->currentPositions[i] = currentPosition[j];
                            contentG->mask[i] = true;
                            break;
                        }
                    }
                    if ( j >= lThis->nbJoints ) {
                        contentG->mask[i] = false;
                    }
                }
            }
            data->motionFinishedCond.notify_one();

        }
    }
    catch (const std::exception &error) {
        std::cerr << "NaoMotionDriver stopped reason: " << error.what() << std::endl;
        data->motionFinishedCond.notify_one();
    }
    pthread_exit(NULL);
}

// Thread =====================================================================
//Frame grabbing function.
void *NaoMotionDriver::xPostureHandlerThread(void *pArg)
{
    //The calling object.
    NaoMotionDriver *lThis = (NaoMotionDriver *)pArg;

    shared_memory_object::remove(shared_memory_posture_name.c_str());
    //Create a shared memory object.
    shared_memory_object shm (open_or_create, shared_memory_posture_name.c_str(), read_write);

    //Set size
    shm.truncate(sizeof(PostureMessage));

    //Map the whole shared memory in this process
    mapped_region region(shm, read_write);

    void *addr = region.get_address();
    //Construct the shared structure in memory
    PostureMessage *data = new (addr) PostureMessage;

    try {
        while (1) {
            scoped_lock<interprocess_mutex> lock(data->newCommandMutex);
            data->newCommandCond.wait(lock);

            std::cerr << "Event Received!" << std::endl;

            float pSpeed = data->speed;
            std::string pName("");

            for (int i = 0; i < 11; i++) {
                if (data->type[i] == '\0') {
                    break;
                }
                pName += data->type[i];
            }

            lThis->setPosture(pName, pSpeed);
            data->motionFinishedCond.notify_one();

        }
    }
    catch (const std::exception &error) {
        std::cerr << "NaoMotionDriver stopped reason: " << error.what() << std::endl;
        data->motionFinishedCond.notify_one();
    }
    pthread_exit(NULL);
}

// Thread =====================================================================
//Frame grabbing function.
void *NaoMotionDriver::xMoveCommandHandlerThread(void *pArg)
{
    //The calling object.
    NaoMotionDriver *lThis = (NaoMotionDriver *)pArg;

    shared_memory_object::remove(shared_memory_move_name.c_str());
    //Create a shared memory object.
    shared_memory_object shm (open_or_create, shared_memory_move_name.c_str(), read_write);

    //Set size
    shm.truncate(sizeof(MoveMessage));

    //Map the whole shared memory in this process
    mapped_region region(shm, read_write);

    void *addr = region.get_address();
    //Construct the shared structure in memory
    MoveMessage *data = new (addr) MoveMessage;


    try {
        while (1) {
            scoped_lock<interprocess_mutex> lock(data->newCommandMutex);
            data->newCommandCond.wait(lock);

            std::cerr << "Event Received!" << std::endl;

            float xp = data->xPos;
            float yp = data->yPos;
            float tp = data->theta;

            if (strcmp(lThis->getPosture().c_str(), "StandInit") != 0 ) {
                lThis->setPosture("StandInit", 0.5f);
            }

            lThis->moveTo(xp, yp, tp);


            data->motionFinishedCond.notify_one();

        }
    }
    catch (const std::exception &error) {
        std::cerr << "NaoMotionDriver stopped reason: " << error.what() << std::endl;
        data->motionFinishedCond.notify_one();
    }
    pthread_exit(NULL);
}
