/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Fabien Badeig, Quentin Pelorson
 * Contacts :
 * perception-naolab@inria.fr
 *
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
* @file naomotiondriver.h
* @author Fabien Badeig
* @brief Driver to manage remotely Nao servo-motors
**/

#ifndef NAOMOTIONDRIVER_H
#define NAOMOTIONDRIVER_H

#include <string>
#include <pthread.h>
#include <boost/interprocess/sync/scoped_lock.hpp>
#include <boost/interprocess/shared_memory_object.hpp>
#include <boost/interprocess/mapped_region.hpp>
#include <boost/interprocess/sync/interprocess_mutex.hpp>
#include <boost/interprocess/sync/interprocess_condition.hpp>

#ifndef Q_MOC_RUN
#include <alcommon/almodule.h>
#include <alcommon/alproxy.h>
#include <alproxies/almotionproxy.h>
#include <alproxies/alrobotpostureproxy.h>
#endif // Q_MOC_RUN

namespace AL
{
class ALBroker;
}

struct NaoPoint {
    float x;
    float y;
    float theta;
};

/**
 * Remote module to manage Nao servo-motors.
 */
class NaoMotionDriver
{
public:
    /**
     * Default Constructor for modules.
     * @param broker the broker to which the module should register.
     */
    NaoMotionDriver(boost::shared_ptr<AL::ALBroker> broker);

    /** Destructor.**/
    ~NaoMotionDriver();

    /**
    * Set the broker to use.
    * @param broker the broker to which the module should register.
    * @return TRUE if connection via broker is successful, FALSE otherwise.
    **/
    bool setMotionProxy(boost::shared_ptr<AL::ALBroker> broker);

    /**
     * Enable stiffness of the specified Nao motors.
     */
    void wakeUp();

    /**
    * Set the stiffness for specific motors.
    * @param state of the stiffness (true = 1.0f and false = 0.0f).
    **/
    void setStiffness( bool lockJoints);

    /**
     * Disable stiffness of the specified Nao motors.
     */
    void rest();

    bool setPostureProxy(boost::shared_ptr<AL::ALBroker> broker);
    void setPosture(std::string name, float speed);

    /**
     * Check if Nao is wakup.
     * @return TRUE if nao is wakeup, FALSE otherwise.
    **/
    bool isWakeUp();

    /**
     * Check if Nao is rest.
     * @return TRUE if nao is sleeping, FALSE otherwise.
    **/
    bool IsRest();

    /**
    * Set the array of motor id.
    * @param pointer on the id array
    * @param array size
    **/
    void setJointNames(int *joints, int nbJoints);

    /**
    * Control specified servo-motor to move according to angle and speed command
    * @param angle to apply for each specified motor
    * @param time to compute each angle
    * @param array size (number of angles to apply)
    * @return TRUE if move is successful, FALSE otherwise.
    **/
    bool moveTo(float *jointAngles, float *jointSpeeds, int nbE);

    void moveTo(float x, float y, float theta);
    void moveTo(NaoPoint toPoint);
    void moveTo(std::vector<NaoPoint> &toPoint);

    /**
    * Check if the new command deal with the same servo-motor
    * @param mask where motor ids are extracted
    * @param number of motor commands (size of the id array)
    * @return TRUE if the servo-motors to control are similar, FALSE otherwise.
    **/
    bool checkJoints(bool *maskID, int nbE);
    std::string getPosture();

    /**
    * Get info about all motor positions
    * @return list of each motor position
    **/
    std::vector<float> getAllJointInfos();

    /**
    * Get info about specified motor position
    * @return list of each specified motor position
    **/
    std::vector<float> getJointInfos();

    /**
    * * Get info about all motor states
    * @return string of motor state (stiffness, control command, real position)
    **/
    std::string getAllInfos();

private:

    /** Set the ALValue (array) of joint names from NAO_control_table **/
    void BuildJointsName();

    /** Start Nao Motion. **/
    void xStartMotionCommandHandler();

    /** Stop Nao Motion. **/
    void xStopMotionCommandHandler();

    /** Start Nao Motion. **/
    void xStartMoveCommandHandler();

    /** Stop Nao Motion. **/
    void xStopMoveCommandHandler();

    /** Start Nao posture thread. **/
    void xStartPostureHandler();

    /** Stop Nao posture thread. **/
    void xStopPostureHandler();

    /** Launch motion control loop.
     * @param pArg user data, here the module instance.
     * @return Thread return value
     **/
    static void *xMotionCommandHandlerThread(void *pArg);
    static void *xMoveCommandHandlerThread(void *pArg);
    static void *xPostureHandlerThread(void *pArg);

    /** Proxy to Motion device. **/
    AL::ALMotionProxy *motionProxy;

    AL::ALRobotPostureProxy *postureProxy;

    /** Motion thread object. **/
    pthread_t fMotionCommandHandlerThread;
    /** Motion thread object. **/
    pthread_t fMoveCommandHandlerThread;
    /** Posture thread object. **/
    pthread_t fPostureHandlerThread;
    /** Number of activated Nao motors**/
    int nbJoints;
    /** Array of joint ID **/
    int *jointID;
    /** array with all joint names **/
    AL::ALValue fJoinsNamesAll;
    /** Selected array of joint names. **/
    AL::ALValue fJoinsNames;
    /** array of requested motor position **/
    AL::ALValue fAngleLists;
    /** array of requested time to compute motor angle **/
    AL::ALValue fTimeLists;

    bool fIsAbsolute;
    bool fMotionThreadRunning;
    bool fMoveThreadRunning;
    bool fPostureThreadRunning;
};

#endif  // NAODMOTIONDRIVER_H

