/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Quentin Perlorson, Fabien Badeig
 * Contacts :
 * perception-naolab@inria.fr
 *
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
 * @file naovideodriver.cpp
 * @author Pelorson Quentin, Fabien Badeig
 * @date 2015 Feb 15
 * @brief
 **/

#include "naovideodriver.h"

#include "commondefinitions.h"

using namespace boost::interprocess;

/**
 *	Pattern :
 * raw[2*i+0]-raw[2*i+1]-raw[2*i+2]-raw[2*i+3]
 * Y - V - Y' - U
 * pixel[i] = YUV;
 * pixel[i+1] = Y'UV;
 **/
void deinterlaceToYuvPixel(unsigned char *rawDataPtr, unsigned char *dataOut)
{
    dataOut[0] = rawDataPtr[0];
    dataOut[1] = rawDataPtr[3];
    dataOut[2] = rawDataPtr[1];
    dataOut[3 + 0] = rawDataPtr[2];
    dataOut[3 + 1] = rawDataPtr[3];
    dataOut[3 + 2] = rawDataPtr[1];
}

void yuvToRgb(unsigned char *dataPtr, int width, int height)
{
    float RCoeff = 1.403f;
    float GCoeff1 = 0.714f;
    float GCoeff2 = 0.344f;
    float BCoeff = 1.773f;
    float delta = 128.0f;

    for (int i = 0 ; i < width * height ; ++i) {
        float R1 = (float(dataPtr[3 * i + 0]) + RCoeff * (float(dataPtr[3 * i + 1]) - delta));
        float G1 = (float(dataPtr[3 * i + 0]) - GCoeff1 * (float(dataPtr[3 * i + 1]) - delta) - GCoeff2 * (float(dataPtr[3 * i + 2]) - delta));
        float B1 = (float(dataPtr[3 * i + 0]) + BCoeff * (float(dataPtr[3 * i + 2]) - delta));

        if (R1 < 0) {
            R1 = 0;
        }
        else if (R1 > 255) {
            R1 = 255;
        }

        if (G1 < 0) {
            G1 = 0;
        }
        else if (G1 > 255) {
            G1 = 255;
        }

        if (B1 < 0) {
            B1 = 0;
        }
        else if (B1 > 255) {
            B1 = 255;
        }

        dataPtr[3 * i + 0] = (unsigned char)R1;
        dataPtr[3 * i + 1] = (unsigned char)G1;
        dataPtr[3 * i + 2] = (unsigned char)B1;
    }
}

void yuvToBgr(unsigned char *dataPtr, int width, int height)
{
    float RCoeff = 1.403f;
    float GCoeff1 = 0.714f;
    float GCoeff2 = 0.344f;
    float BCoeff = 1.773f;
    float delta = 128.0f;

    for (int i = 0 ; i < width * height ; ++i) {
        float R1 = (float(dataPtr[3 * i + 0]) + RCoeff * (float(dataPtr[3 * i + 1]) - delta));
        float G1 = (float(dataPtr[3 * i + 0]) - GCoeff1 * (float(dataPtr[3 * i + 1]) - delta) - GCoeff2 * (float(dataPtr[3 * i + 2]) - delta));
        float B1 = (float(dataPtr[3 * i + 0]) + BCoeff * (float(dataPtr[3 * i + 2]) - delta));

        if (R1 < 0) {
            R1 = 0;
        }
        else if (R1 > 255) {
            R1 = 255;
        }

        if (G1 < 0) {
            G1 = 0;
        }
        else if (G1 > 255) {
            G1 = 255;
        }

        if (B1 < 0) {
            B1 = 0;
        }
        else if (B1 > 255) {
            B1 = 255;
        }

        dataPtr[3 * i + 0] = (unsigned char)B1;
        dataPtr[3 * i + 1] = (unsigned char)G1;
        dataPtr[3 * i + 2] = (unsigned char)R1;
    }
}

// NaoVideoDriver Class ==========================================================
/**
 * Constructor for NaoVideoDriver object
 * @param broker The parent broker
 * @param name The name of the module
 */
NaoVideoDriver::NaoVideoDriver(boost::shared_ptr<AL::ALBroker> broker, std::string type):
    fCapturingVideo(false)
    , fALVideoDevice(0)
    , fVideoSubscriberId()
    , fVideoThread()
    , fFramerate(0.0f)
    , fType(type)
{
    if (broker) {
        fALVideoDevice = new AL::ALVideoDeviceProxy(broker);
    }

    if ( fType.compare("mono") == 0 ) {
        shared_memory_object::remove(shared_memory_video0_name.c_str());
        //Create a shared memory object.
        shared_memory_object shm0 (create_only, shared_memory_video0_name.c_str(), read_write);
        //Set size
        shm0.truncate(sizeof(ImageMessage));
    }
    else if ( fType.compare("stereo") == 0 ) {
        shared_memory_object::remove(shared_memory_video0_name.c_str());
        shared_memory_object::remove(shared_memory_video1_name.c_str());
        //Create a shared memory object.
        shared_memory_object shm0 (create_only, shared_memory_video0_name.c_str(), read_write);
        shared_memory_object shm1 (create_only, shared_memory_video1_name.c_str(), read_write);

        //Set size
        shm0.truncate(sizeof(ImageMessage));
        shm1.truncate(sizeof(ImageMessage));
    }
    else {
        std::cerr << "Error with stream type (mono/stereo)" << std::endl;
        exit(EXIT_FAILURE);
    }


    pthread_mutex_init(&fReadyMutex, NULL);
    pthread_cond_init (&fThreadReadySignal, NULL);
}

NaoVideoDriver::~NaoVideoDriver()
{
    stopCapture();
    delete fALVideoDevice;
    if ( fType.compare("mono") == 0 ) {
        shared_memory_object::remove(shared_memory_video0_name.c_str());
    }
    else if ( fType.compare("stereo") == 0 ) {
        shared_memory_object::remove(shared_memory_video0_name.c_str());
        shared_memory_object::remove(shared_memory_video1_name.c_str());
    }
    pthread_mutex_destroy(&fReadyMutex);
    pthread_cond_destroy(&fThreadReadySignal);
}

bool NaoVideoDriver::setBroker(boost::shared_ptr<AL::ALBroker> broker)
{
    if (fALVideoDevice) {
        if (isConnected()) {
            xStopVideo();
        }
        delete fALVideoDevice;
    }
    if (broker) {
        fALVideoDevice = new AL::ALVideoDeviceProxy(broker);
    }
    else {
        fALVideoDevice = 0;
    }
    return fALVideoDevice != 0;
}

void NaoVideoDriver::initializeCam()
{
    if ( fType.compare("mono") == 0 ) {
        cams.arraySetSize(1);
        cams[0] = AL::kTopCamera;
        desiredSizes.arraySetSize(1);
        desiredColorSpaces.arraySetSize(1);
        desiredSizes[0] = AL::kVGA;
        desiredColorSpaces[0] = AL::kYuvColorSpace;
        //desiredColorSpaces[0] = AL::kRGBColorSpace;
    }
    else if ( fType.compare("stereo") == 0 ) {
        cams.arraySetSize(2);
        cams[0] = AL::kTopCamera;
        cams[1] = AL::kBottomCamera;
        desiredSizes.arraySetSize(2);
        desiredColorSpaces.arraySetSize(2);
        for (int i = 0; i < 2; i++) {
            desiredSizes[i] = AL::kVGA;
            //desiredColorSpaces[i] = AL::kBGRColorSpace;
            desiredColorSpaces[i] = AL::kYuvColorSpace;
        }
    }
}

void NaoVideoDriver::startCapture()
{
    try {
        if (!fALVideoDevice) {
            throw std::runtime_error("Need to create a valid broker before.");
            return;
        }

        if (fCapturingVideo) {
            throw std::runtime_error("Capture already started, can't start a second one.");
        }
        else {
            xStartVideo();
            std::cout << "STREAMING STARTED." << std::endl;
        }
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot start : " << error.what() << std::endl;
    }
}

void NaoVideoDriver::stopCapture()
{
    if (fCapturingVideo) {
        xStopVideo();
    }
}

bool NaoVideoDriver::isConnected()
{
    return fCapturingVideo;
}


float NaoVideoDriver::getFramerate()
{
    return fFramerate;
}
// Video ======================================================================
void NaoVideoDriver::xStartVideo()
{
    try {
        if ( fType.compare("mono") == 0 ) {
            fVideoSubscriberId = fALVideoDevice->subscribeCamera("NaoVideoDriver", cams[0], desiredSizes[0], desiredColorSpaces[0], 15);
        }
        else if ( fType.compare("stereo") == 0 ) {
            fVideoSubscriberId = fALVideoDevice->subscribeCameras("NaoVideoDriver", cams, desiredSizes, desiredColorSpaces, 15);
        }
        fCapturingVideo = true;
        pthread_create(&fVideoThread, 0, xVideoThread, this);
        std::cout << "Capture thread created." << std::endl;
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot subscribe to video: " << error.what() << std::endl;
        fCapturingVideo = false;
    }
}

void NaoVideoDriver::xStopVideo()
{
    void *lReturn;
    fCapturingVideo = false;
    pthread_join(fVideoThread, &lReturn);
    try {
        fALVideoDevice->unsubscribe(fVideoSubscriberId);
    }
    catch (const std::exception &error) {
        std::cerr << "Cannot unsubscribe from video: " << error.what() << std::endl;
    }
    std::cout << "Video Thread closed & Module unsubscribed" << std::endl;
}

// Thread =====================================================================
//Frame grabbing function.
void *NaoVideoDriver::xVideoThread(void *pArg)
{
    //The calling object.
    NaoVideoDriver *lThis = (NaoVideoDriver *)pArg;
    // Time measurement.
    long long previoustime = xGetTime();

    ImageMessage *data0;
    ImageMessage *data1;
    shared_memory_object shm0;
    shared_memory_object shm1;
    mapped_region region0;
    mapped_region region1;
    void *addr0;
    void *addr1;

#if 0
    char filename[256];
    cv::Mat imgCV(cv::Size(640, 480), CV_8UC3);
#endif

    if ( lThis->fType.compare("mono") == 0 ) {
        //Create a shared memory object.
        shm0 = shared_memory_object(open_only, shared_memory_video0_name.c_str(), read_write);

        //Map the whole shared memory in this process
        region0 = mapped_region(shm0, read_write);

        pthread_mutex_lock(&lThis->fReadyMutex);
        pthread_cond_signal(&lThis->fThreadReadySignal);
        pthread_mutex_unlock(&lThis->fReadyMutex);

        addr0 = region0.get_address();
        //Construct the shared structure in memory
        data0 = new (addr0) ImageMessage;

    }
    else if ( lThis->fType.compare("stereo") == 0 ) {
        //Create a shared memory object.
        shm0 = shared_memory_object(open_only, shared_memory_video0_name.c_str(), read_write);
        shm1 = shared_memory_object(open_only, shared_memory_video1_name.c_str(), read_write);

        //Map the whole shared memory in this process
        region0 = mapped_region(shm0, read_write);
        region1 = mapped_region(shm1, read_write);

        //pthread_mutex_lock(&lThis->fReadyMutex);
        //pthread_cond_signal(&lThis->fThreadReadySignal);
        //pthread_mutex_unlock(&lThis->fReadyMutex);

        addr0 = region0.get_address();
        //Construct the shared structure in memory
        data0 = new (addr0) ImageMessage;

        addr1 = region1.get_address();
        //Construct the shared structure in memory
        data1 = new (addr1) ImageMessage;
    }

    try {
        while (lThis->fCapturingVideo) {
            //Compute the framerate.
            long long currenttime = xGetTime();
            float frameDuration =  (currenttime - previoustime) / 1000.0f;
            float fps = 1000.0f / frameDuration;
            previoustime = currenttime;
            lThis->fFramerate = fps;
            //Get image from nao through the broker.
            if ( lThis->fType.compare("mono") == 0 ) {
                lThis->fRawImages = lThis->fALVideoDevice->getDirectRawImageRemote(lThis->fVideoSubscriberId);
                if ( lThis->fRawImages.getSize() == 0 ) {
                    lThis->fALVideoDevice->releaseDirectRawImage(lThis->fVideoSubscriberId);
                    continue;
                }

                {
                    boost::interprocess::scoped_lock<boost::interprocess::interprocess_upgradable_mutex> lock(data0->imageMutex);
                    data0->ts = lThis->fRawImages[4];
                    data0->tms = lThis->fRawImages[5];
                    data0->width = lThis->fRawImages[0];
                    data0->height = lThis->fRawImages[1];
                    data0->nbChannels = 3;

                    unsigned char *rawDataPtr0 = (unsigned char *)lThis->fRawImages[6].GetBinary();

                    int width0 = (int) lThis->fRawImages[0];
                    int height0 = (int) lThis->fRawImages[1];

                    unsigned char *sharedBufferPtr0 = data0->items;

                    for (int i = 0 ; i < width0 * height0 ; i = i + 2) {
                        deinterlaceToYuvPixel(rawDataPtr0 + 2 * i, sharedBufferPtr0 + 3 * i);
                    }
                    yuvToRgb(sharedBufferPtr0, width0, height0);
                    std::cout << "[MONO] " << lThis->fRawImages[4] << "," << lThis->fRawImages[5] << ": Grabbing image -- FPS: " << fps << std::endl;
                    data0->imageCond.notify_all();
                }
                lThis->fALVideoDevice->releaseDirectRawImage(lThis->fVideoSubscriberId);
            }
            else if ( lThis->fType.compare("stereo") == 0 ) {
                lThis->fRawImages = lThis->fALVideoDevice->getDirectRawImagesRemote(lThis->fVideoSubscriberId);

                //Write frame in the shared memory.
                {
                    boost::interprocess::scoped_lock<boost::interprocess::interprocess_upgradable_mutex> lock(data0->imageMutex);
                    std::cout << "[STEREO] " << std::endl;

                    if ( data0->tms != (int)lThis->fRawImages[0][5]) {
                        data0->ts = lThis->fRawImages[0][4];
                        data0->tms = lThis->fRawImages[0][5];
                        data0->width = lThis->fRawImages[0][0];
                        data0->height = lThis->fRawImages[0][1];
                        data0->nbChannels = 3;

                        unsigned char *rawDataPtr0 = (unsigned char *)lThis->fRawImages[0][6].GetBinary();

                        int width0 = (int) lThis->fRawImages[0][0];
                        int height0 = (int) lThis->fRawImages[0][1];

                        unsigned char *sharedBufferPtr0 = data0->items;

                        for (int i = 0 ; i < width0 * height0 ; i = i + 2) {
                            deinterlaceToYuvPixel(rawDataPtr0 + 2 * i, sharedBufferPtr0 + 3 * i);
                        }
                        yuvToRgb(sharedBufferPtr0, width0, height0);
                        std::cout << lThis->fRawImages[0][4] << "," << lThis->fRawImages[0][5] << ": Grabbing right image" << std::endl;
                    }
                    data0->imageCond.notify_all();
                }

                {
                    boost::interprocess::scoped_lock<boost::interprocess::interprocess_upgradable_mutex> lock(data1->imageMutex);
                    if ( data1->tms != (int)lThis->fRawImages[1][5]) {
                        data1->ts = lThis->fRawImages[1][4];
                        data1->tms = lThis->fRawImages[1][5];
                        data1->width = lThis->fRawImages[1][0];
                        data1->height = lThis->fRawImages[1][1];
                        data1->nbChannels = 3;

                        int width1 = (int) lThis->fRawImages[1][0];
                        int height1 = (int) lThis->fRawImages[1][1];

                        unsigned char *rawDataPtr1 = (unsigned char *)lThis->fRawImages[1][6].GetBinary();

                        unsigned char *sharedBufferPtr1 = data1->items;

                        for (int i = 0 ; i < width1 * height1 ; i = i + 2) {
                            deinterlaceToYuvPixel(rawDataPtr1 + 2 * i, sharedBufferPtr1 + 3 * i);
                        }
                        yuvToRgb(sharedBufferPtr1, width1, height1);
                        std::cout << lThis->fRawImages[1][4] << "," << lThis->fRawImages[1][5] << ": Grabbing left image" << std::endl;
                    }
                    data1->imageCond.notify_all();
                }
                std::cout << " -- Timestamp delay: " <<  (int)lThis->fRawImages[1][5] - (int)lThis->fRawImages[0][5] << " --" << std::endl;
                //Releasing the image for video device.
                lThis->fALVideoDevice->releaseDirectRawImages(lThis->fVideoSubscriberId);
            }
        }
    }
    catch (const std::exception &error) {
        std::cerr << "Video loop stopped, reason: " << error.what() << std::endl;
        if (lThis->fCapturingVideo) {
            lThis->xStopVideo();
        }
    }
    pthread_exit(NULL);
}

long long NaoVideoDriver::xGetTime()
{
    qi::os::timeval lTimeStruct;
    qi::os::gettimeofday(&lTimeStruct);
    long long lTime = (long long)lTimeStruct.tv_sec * 1000000; // Seconds
    lTime += (long long)lTimeStruct.tv_usec;                   // Microseconds
    return lTime;
}
