/* ============================================================
 *
 * This file is a part of the EARS project
 * Copyright (c) INRIA 2015
 *
 * Contributor (s) : Quentin Perlorson, Fabien Badeig
 * Contacts :
 * perception-naolab@inria.fr
 *
 * This file is part of NaoLab.
 *
 * NaoLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * NaoLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with NaoLab. If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * This project is headed by Perception research team at INRIA.
 *
 * ============================================================*/

/**
* @file naovideodriver.h
* @author Pelorson Quentin, Fabien Badeig
* @date 2015 Feb 15
* @brief
**/

#ifndef NAOVIDEODRIVER_H
#define NAOVIDEODRIVER_H

#include <string>
#include <pthread.h>
#include <semaphore.h>
#include <boost/interprocess/sync/scoped_lock.hpp>
#include <boost/interprocess/shared_memory_object.hpp>
#include <boost/interprocess/mapped_region.hpp>
#include <boost/interprocess/sync/interprocess_mutex.hpp>
#include <boost/interprocess/sync/interprocess_condition.hpp>

#ifndef Q_MOC_RUN
#include <alcommon/almodule.h>
#include <alcommon/alproxy.h>
#include <alvision/alvisiondefinitions.h>
#include <alproxies/alvideodeviceproxy.h>
#endif // Q_MOC_RUN

namespace AL
{
class ALBroker;
}

/**
 * Remote module for synchronized audio and video capture.
 */
class NaoVideoDriver
{
public:
    /**
     * Default Constructor for modules.
     * @param broker the broker to which the module should register.
     */
    NaoVideoDriver(boost::shared_ptr<AL::ALBroker> broker, std::string type);

    /** Destructor.**/
    ~NaoVideoDriver();

    /**
    * Set the broker to use.
    * @param broker the broker to which the module should register.
    * @return TRUE if connection via broker is successful, FALSE otherwise.
    **/
    bool setBroker(boost::shared_ptr<AL::ALBroker> broker);

    void initializeCam();

    /**
     * Start capture.
     * @param camResolution the video resolution see CamResolution.
     */
    void startCapture();

    /** Stop capture.**/
    void stopCapture();

    /**
     * Return the state of the connection to the video device on nao.
     * @return TRUE if connected to nao, FALSE otherwise.
    **/
    bool isConnected();

    /**
     * Return the frame receiving framerate .
     * @return the framerate in Frame/second.
    **/
    float getFramerate();
private:
    /** Start video capture. **/
    void xStartVideo();

    /** Stop video capture. **/
    void xStopVideo();

    /** Video acquisition loop.
     * @param pArg user data, here the module instance.
     * @return Thread return value
     **/
    static void *xVideoThread(void *pArg);

    /** Get the time in Us
     * @return time in Us
    **/
    static long long xGetTime();

    /** Are we currently capturing video ? **/
    bool fCapturingVideo;

    /** Proxy to video device. **/
    AL::ALVideoDeviceProxy *fALVideoDevice;

    /** Video subscriber id. **/
    std::string fVideoSubscriberId;

    /** Video thread object. **/
    pthread_t fVideoThread;

    /** Threads synchronization mutex. **/
    pthread_mutex_t fReadyMutex;

    /** Threads synchronization signal.**/
    pthread_cond_t fThreadReadySignal;

    AL::ALValue cams;
    AL::ALValue desiredSizes;
    AL::ALValue desiredColorSpaces;
    /** Raw Image Buffer. **/
    AL::ALValue fRawImages;

    /** Video framerate. **/
    float fFramerate;
    std::string fType;
};

#endif  // NAOVIDEODRIVER_H

